# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QCheckBox, QComboBox

from ...helpers import BaseDialog
from ...shell import EventType

from ..helpers import tagged_items

from .helpers import init_feature_selector


class DeleteFeatureDialog(BaseDialog):
    """ This class implements the dialog for deleting a feature. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        self._feature = QComboBox()
        layout.addWidget(self._feature)

        self._discard = QCheckBox(
                "Discard feature-specific parts of the project?")
        layout.addWidget(self._discard)

    def set_fields(self):
        """ Set the dialog's fields from the project. """

        init_feature_selector(self._feature, self.model)

    def get_fields(self):
        """ Update the project from the dialog's fields. """

        project = self.model

        feature = self._feature.currentText()
        discard = self._discard.isChecked()

        # Delete from each API item it appears.
        remove_items = []

        for api_item, container in tagged_items(project):
            # Ignore items that aren't tagged with a feature.
            if len(api_item.features) == 0:
                continue

            remove_features = []

            for f in api_item.features:
                if f[0] == '!':
                    if f[1:] == feature:
                        # Just remove it if it is not the only one or if we are
                        # discarding if enabled (and the feature is inverted).
                        if len(api_item.features) > 1 or discard:
                            remove_features.append(f)
                        else:
                            remove_items.append((api_item, container))
                            break

                elif f == feature:
                    # Just remove it if it is not the only one or if we are
                    # discarding if enabled.
                    if len(api_item.features) > 1 or not discard:
                        remove_features.append(f)
                    else:
                        remove_items.append((api_item, container))
                        break
            else:
                # Note that we deal with a feature appearing multiple times,
                # even though that is probably a user bug.
                for f in remove_features:
                    api_item.features.remove(f)

        for api_item, container_item in remove_items:
            container_item.content.remove(api_item)
            self.shell.notify(EventType.CONTAINER_API_DELETE,
                    (container_item, api_item))

        # Delete from the project's list.
        if feature in project.externalfeatures:
            feature_list = project.externalfeatures
        else:
            feature_list = project.features

        feature_list.remove(feature)

        self.shell.notify(EventType.FEATURE_DELETE, feature)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass, field

from .access import Access
from .code import Code
from .enum_value import EnumValue


@dataclass
class Enum(Code, Access):
    """ This class implements an enum. """

    # The list of enum values.
    content: list[EnumValue] = field(default_factory=list)

    # Set if the enum is a C++11 enum class.
    enumclass: bool = False

    # The optional name of the enum.
    name: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


import os

from PyQt6.QtGui import QAction
from PyQt6.QtWidgets import QFileDialog

from ...models import Project
from ...project_io import load_project

from ..helpers import ProjectUi, warning
from ..shell_tool import ActionTool


class ImportProjectTool(ActionTool):
    """ This class implements a tool importing another project. """

    @property
    def actions(self):
        """ Get the destination menu and sequence of actions handled by the
        tool.
        """

        self._import_action = QAction("Import Project...",
                triggered=self._handle_import)

        return ("Tools", (self._import_action, ))

    @property
    def name(self):
        """ Get the tool's name. """

        return 'metasip.tool.import_project'

    def _handle_import(self):
        """ Handle the Import action. """

        dir_name = os.path.dirname(self.shell.project.name)

        import_name, _ = QFileDialog.getOpenFileName(self.shell.widget,
                "Import project file", dir_name,
                "MetaSIP project files (*.msp)")

        if import_name:
            imported = Project(name=import_name)
            if not load_project(imported, ui=ProjectUi()):
                return

            try:
                self._import_project(imported)
            except Exception as e:
                warning(str(e))

    def _import_project(self, imported):
        """ Import a project. """

        project = self.shell.project

        project_name = self._project_name(project)
        imported_name = self._project_name(imported)

        # Check the project being imported doesn't have any versions defined as
        # we don't support multiple timelines.
        if len(imported.versions) != 0:
            raise Exception(f"'{imported_name}' defines one or more versions")

        # Merge any external features.
        for feature in list(project.externalfeatures):
            if feature in imported.features:
                project.externalfeatures.remove(feature)

        for feature in imported.externalfeatures:
            if feature in project.features:
                continue

            if feature in project.externalfeatures:
                continue

            project.externalfeatures.append(feature)

        # Add any new features and check for conflicts.
        for feature in imported.features:
            if feature in project.features:
                raise Exception(f"Both '{project_name}' and '{imported_name}' define a '{feature}' feature")

        # Merge any externally defined modules.
        for module_name in list(project.externalmodules):
            for module in imported.modules:
                if module.name == module_name:
                    break
            else:
                project.externalmodules.remove(module_name)

        for module_name in imported.externalmodules:
            if module_name in project.externalmodules:
                continue

            for module in project.modules:
                if module.name == module_name:
                    break
            else:
                project.externalmodules.append(module_name)

        # Merge any platforms.
        for platform in imported.platforms:
            if platform not in project.platforms:
                project.platforms.append(platform)

        # Any any new modules and check for conflicts.
        for imported_module in imported.modules:
            for module in project.modules:
                if module.name == imported_module.name:
                    raise Exception(f"Both '{project_name}' and '{imported_name}' define a '{module.name}' module")

            project.modules.append(imported_module)

        # Any any new header directories and check for conflicts.
        for imported_header in imported.headers:
            for header in project.headers:
                if header.name == imported_header.name:
                    raise Exception(f"Both '{project_name}' and '{imported_name}' define a '{header.name}' header directory")

            project.headers.append(imported_headers)

        self.shell.dirty = True

    @staticmethod
    def _project_name(project):
        """ Return a project's name for use in user messages. """

        return os.path.splitext(os.path.basename(project.name))[0]



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass


@dataclass
class Annos:
    """ This class is a mixin for API models that may have SIP annotations. """

    # The annotations.
    annos: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import (QCheckBox, QComboBox, QFormLayout, QGroupBox,
        QLineEdit, QVBoxLayout)

from ....helpers import BaseDialog

from .helpers import Encoding, split_annos


class TypedefPropertiesDialog(BaseDialog):
    """ This class implements the dialog for a typedef's properties. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        group_box = QGroupBox("Annotations")
        layout.addWidget(group_box)
        group_box_layout = QVBoxLayout()
        group_box.setLayout(group_box_layout)

        self._capsule = QCheckBox('Capsule')
        group_box_layout.addWidget(self._capsule)

        form = QFormLayout()
        group_box_layout.addLayout(form)

        encoding = QComboBox()
        self._encoding_helper = Encoding(encoding)
        form.addRow('Encoding', encoding)

        self._no_type_name = QCheckBox('NoTypeName')
        group_box_layout.addWidget(self._no_type_name)

        self._py_int = QCheckBox('PyInt')
        group_box_layout.addWidget(self._py_int)

        form = QFormLayout()
        group_box_layout.addLayout(form)

        self._py_name = QLineEdit()
        form.addRow('PyName', self._py_name)

        self._type_hint = QLineEdit()
        form.addRow('TypeHint', self._type_hint)

        self._type_hint_in = QLineEdit()
        form.addRow('TypeHintIn', self._type_hint_in)

        self._type_hint_out = QLineEdit()
        form.addRow('TypeHintOut', self._type_hint_out)

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        for name, value in split_annos(self.model.annos):
            if name == 'Capsule':
                self._capsule.setChecked(True)
            elif name == 'Encoding':
                self._encoding_helper.setAnnotation(value)
            elif name == 'NoTypeName':
                self._no_type_name.setChecked(True)
            elif name == 'PyInt':
                self._py_int.setChecked(True)
            elif name == 'PyName':
                self._py_name.setText(value)
            elif name == 'TypeHint':
                self._type_hint.setText(value)
            elif name == 'TypeHintIn':
                self._type_hint_in.setText(value)
            elif name == 'TypeHintOut':
                self._type_hint_out.setText(value)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        annos_list = []

        self._encoding_helper.annotation(annos_list)

        if self._capsule.isChecked():
            annos_list.append('Capsule')

        if self._no_type_name.isChecked():
            annos_list.append('NoTypeName')

        if self._py_int.isChecked():
            annos_list.append('PyInt')

        py_name = self._py_name.text().strip()
        if py_name:
            annos_list.append('PyName=' + py_name)

        type_hint = self._type_hint.text().strip()
        if type_hint:
            annos_list.append(f'TypeHint="{type_hint}"')

        type_hint_in = self._type_hint_in.text().strip()
        if type_hint_in:
            annos_list.append(f'TypeHintIn="{type_hint_in}"')

        type_hint_out = self._type_hint_out.text().strip()
        if type_hint_out:
            annos_list.append(f'TypeHintOut="{type_hint_out}"')

        self.model.annos = ','.join(annos_list)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .callable import Callable
from .docstring import Docstring


@dataclass
class Function(Callable, Docstring):
    """ This class implements a global C/C++ function. """



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..header_file_version import HeaderFileVersion

from .adapt import adapt
from .base_adapter import AttributeType, BaseAdapter


class HeaderFileAdapter(BaseAdapter):
    """ This is the HeaderFile adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'ignored':  AttributeType.BOOL,
        'module':   AttributeType.STRING,
        'name':     AttributeType.STRING,
    }

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        super().load(element, project, ui)

        for subelement in element:
            if subelement.tag == 'HeaderFileVersion':
                header_file_version = HeaderFileVersion()
                adapt(header_file_version).load(subelement, project, ui)
                self.model.versions.append(header_file_version)

    def save(self, output):
        """ Save the model to an output file. """

        header_file = self.model

        output.write(f'<HeaderFile name="{header_file.name}"')
        self.save_str('module', output)
        self.save_bool('ignored', output)
        output.write('>\n')
        output += 1

        for header_file_version in header_file.versions:
            adapt(header_file_version).save(output)

        output -= 1
        output.write('</HeaderFile>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..access import Access
from ..annos import Annos
from ..callable import Callable

from .adapt import adapt
from .base_adapter import AttributeType, BaseApiAdapter


class OperatorMethodAdapter(BaseApiAdapter):
    """ This is the OperatorMethod adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'abstract': AttributeType.BOOL,
        'const':    AttributeType.BOOL,
        'virtcode': AttributeType.LITERAL,
        'virtual':  AttributeType.BOOL,
    }

    def __eq__(self, other):
        """ Compare for C/C++ equality. """

        method = self.model
        other_method = other.model

        if type(method) is not type(other_method):
            return False

        if adapt(method, Callable) != adapt(other_method, Callable):
            return False

        if method.access != other_method.access:
            return False

        if method.virtual != other_method.virtual:
            return False

        if method.const != other_method.const:
            return False

        if method.abstract != other_method.abstract:
            return False

        return True

    def as_str(self):
        """ Return the standard string representation. """

        # We can't use the super class version because we might need to stick
        # some text in the middle of it.

        method = self.model

        callable_adapter = adapt(method, Callable)

        s = ''

        if method.virtual:
            s += 'virtual '

        s += callable_adapter.return_type_as_str(allow_py=True) + 'operator' + method.name

        if method.pyargs != '':
            s += method.pyargs
        else:
            args = ', '.join([adapt(arg).as_py_str() for arg in method.args])
            s += '(' + args + ')'

        if method.const:
            s += ' const'

        if method.abstract:
            s += ' = 0'

        s += adapt(method, Annos).as_str()

        if (method.virtual or method.access.startswith('protected') or method.methcode == '') and callable_adapter.has_different_signatures():
            return_type = callable_adapter.return_type_as_str().strip()
            args = ', '.join([adapt(arg).as_str() for arg in method.args])
            s += f' [{return_type} ({args})]'

        return s

    def generate_sip(self, sip_file, output):
        """ Generate the .sip file content. """

        method = self.model

        nr_ends = self.version_start(output)

        output.write(self.as_str())
        output.write(';\n')

        adapt(method, Callable).generate_sip_directives(output)
        adapt(method, Access).generate_sip_directives(output)
        output.write_code_directive('%VirtualCatcherCode', method.virtcode)

        self.version_end(nr_ends, output)

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        super().load(element, project, ui)

        adapt(self.model, Callable).load(element, project, ui)
        adapt(self.model, Access).load(element, project, ui)

    def save(self, output):
        """ Save the model to an output file. """

        method = self.model

        output.write('<OperatorMethod')
        adapt(method, Callable).save_attributes(output)
        adapt(method, Access).save_attributes(output)
        self.save_bool('virtual', output)
        self.save_bool('const', output)
        self.save_bool('abstract', output)
        output.write('>\n')

        output += 1
        adapt(method, Callable).save_subelements(output)
        adapt(method, Access).save_subelements(output)
        self.save_literal('virtcode', output)
        output -= 1

        output.write('</OperatorMethod>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .access import Access
from .code import Code


@dataclass
class OpaqueClass(Code, Access):
    """ This class implements an opaque class. """

    # The name of the class.
    name: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .callable import Callable


@dataclass
class OperatorFunction(Callable):
    """ This class implements a global C++ operator. """



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QFormLayout, QLineEdit

from ....helpers import BaseDialog


class ManualCodeDialog(BaseDialog):
    """ This class implements the dialog for manual code. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        form = QFormLayout()
        layout.addLayout(form)

        self._precis = QLineEdit()
        form.addRow("Precis", self._precis)

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        self._precis.setText(self.model.precis)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        self.model.precis = self._precis.text().strip()

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QCheckBox, QGroupBox, QVBoxLayout

from ....helpers import BaseDialog

from .helpers import split_annos


class OpaqueClassPropertiesDialog(BaseDialog):
    """ This class implements the dialog for an opaque class's properties. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        group_box = QGroupBox("Annotations")
        group_box_layout = QVBoxLayout()
        group_box.setLayout(group_box_layout)
        layout.addWidget(group_box)

        self._external = QCheckBox('External')
        group_box_layout.addWidget(self._external)

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        for name, value in split_annos(self.model.annos):
            if name == 'External':
                self._external.setChecked(True)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        annos_list = []

        if self._external.isChecked():
            annos_list.append('External')

        self.model.annos = ','.join(annos_list)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..annos import Annos
from ..code import Code
from ..docstring import Docstring

from .adapt import adapt
from .base_adapter import AttributeType, BaseApiAdapter


class TypedefAdapter(BaseApiAdapter):
    """ This is the Typedef adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'name': AttributeType.STRING,
        'type': AttributeType.STRING,
    }

    def __eq__(self, other):
        """ Compare for C/C++ equality. """

        typedef = self.model
        other_typedef = other.model

        if type(typedef) is not type(other_typedef):
            return False

        if typedef.name != other_typedef.name:
            return False

        if self.expand_type(typedef.type) != self.expand_type(other_typedef.type):
            return False

        return True

    def as_str(self):
        """ Return the standard string representation. """

        typedef = self.model

        return 'typedef ' + self.expand_type(typedef.type, typedef.name) + adapt(typedef, Annos).as_str()

    def generate_sip(self, sip_file, output):
        """ Generate the .sip file content. """

        typedef = self.model

        nr_ends = self.version_start(output)

        output.write(self.as_str())
        output.write(';\n')

        adapt(typedef, Docstring).generate_sip_directives(output)

        self.version_end(nr_ends, output)

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        super().load(element, project, ui)

        adapt(self.model, Code).load(element, project, ui)
        adapt(self.model, Docstring).load(element, project, ui)

    def save(self, output):
        """ Save the model to an output file. """

        typedef = self.model

        output.write('<Typedef')
        adapt(typedef, Code).save_attributes(output)
        adapt(typedef, Docstring).save_attributes(output)
        self.save_attribute('name', typedef.name, output)
        self.save_attribute('type', typedef.type, output)
        output.write('>\n')

        output += 1
        adapt(typedef, Code).save_subelements(output)
        adapt(typedef, Docstring).save_subelements(output)
        output -= 1

        output.write('</Typedef>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from xml.etree import ElementTree

from ..exceptions import UserException
from ..models import MinimumProjectVersion, ProjectVersion
from ..models.adapters import adapt


def load_project(project, ui=None):
    """ Populate a project from its project file.  Return True if the user
    didn't cancel.
    """

    # Load the file.
    tree = ElementTree.parse(project.name)

    # Do some basic sanity checks.
    root = tree.getroot()

    major_version = root.get('majorversion')
    minor_version = root.get('minorversion')

    if major_version is None and minor_version is None:
        major_version = 0
        minor_version = _as_int(root.get('version'))
    else:
        major_version = _as_int(major_version)
        minor_version = _as_int(minor_version)

    if root.tag != 'Project' or major_version < 0 or minor_version < 0:
        raise UserException(
                f"{project.name} doesn't appear to be a valid metasip project")

    # Check we handle the version.
    version = (major_version, minor_version)

    if version < MinimumProjectVersion:
        raise UserException(
                f"{project.name} was created with an unsupported version of metasip")

    if version > ProjectVersion:
        raise UserException(
                f"{project.name} was created with a later version of metasip")

    # See if user input is required.
    if version[0] != ProjectVersion[0]:
        if ui is None:
            raise UserException(
                    f"{project.name} was created with an earlier version of metasip and must be updated using the GUI")

        if not ui.update_project_format(root, version, ProjectVersion):
            return False

        project.dirty = True

    elif version[1] != ProjectVersion[1]:
        if ui is not None:
            ui.warn_minor_version_update(version, ProjectVersion)

        project.version = version

    # Populate the project.
    adapt(project).load(root, project, ui)

    return True


def _as_int(s):
    """ Return an int from a string or -1 if the string is invalid. """

    try:
        return int(s)
    except ValueError:
        return -1



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from .base_adapter import AttributeType, BaseAdapter


class HeaderFileVersionAdapter(BaseAdapter):
    """ This is the HeaderFileVersion adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'md5':      AttributeType.STRING,
        'parse':    AttributeType.BOOL,
        'version':  AttributeType.STRING,
    }

    def save(self, output):
        """ Save the model to an output file. """

        header_file_version = self.model

        output.write(f'<HeaderFileVersion md5="{header_file_version.md5}" version="{header_file_version.version}"')
        self.save_bool('parse', output)
        output.write('/>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from .adapt import adapt
from .base_adapter import BaseAdapter


class CodeContainerAdapter(BaseAdapter):
    """ This is the CodeContainer adapter. """

    def load(self, tag_code_map, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        for subelement in element:
            model_factory = tag_code_map.get(subelement.tag)
            if model_factory is not None:
                model = model_factory()
                adapt(model).load(subelement, project, ui)
                self.model.content.append(model)

    def save_subelements(self, output):
        """ Save the XML subelements. """

        for code in self.model.content:
            adapt(code).save(output)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import (QCheckBox, QFormLayout, QGridLayout, QGroupBox,
        QLineEdit)

from ....helpers import BaseDialog

from .helpers import split_annos


class ClassPropertiesDialog(BaseDialog):
    """ This class implements the dialog for a class's properties. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        form = QFormLayout()
        layout.addLayout(form)

        self._py_bases = QLineEdit()
        form.addRow("Python base classes", self._py_bases)

        group_box = QGroupBox("Annotations")
        group_box_layout = QGridLayout()
        group_box.setLayout(group_box_layout)
        layout.addWidget(group_box)

        self._abstract = QCheckBox('Abstract')
        group_box_layout.addWidget(self._abstract, 0, 0)

        self._allow_none = QCheckBox('AllowNone')
        group_box_layout.addWidget(self._allow_none, 1, 0)

        form = QFormLayout()
        group_box_layout.addLayout(form, 2, 0)

        self._api = QLineEdit()
        form.addRow('API', self._api)

        self._delay_dtor = QCheckBox('DelayDtor')
        group_box_layout.addWidget(self._delay_dtor, 3, 0)

        self._export_derived = QCheckBox('ExportDerived')
        group_box_layout.addWidget(self._export_derived, 4, 0)

        form = QFormLayout()
        group_box_layout.addLayout(form, 5, 0)

        self._metatype = QLineEdit()
        form.addRow('Metatype', self._metatype)

        self._mixin = QCheckBox('Mixin')
        group_box_layout.addWidget(self._mixin, 6, 0)

        self._no_default_ctors = QCheckBox('NoDefaultCtors')
        group_box_layout.addWidget(self._no_default_ctors, 7, 0)

        self._no_type_hint = QCheckBox('NoTypeHint')
        group_box_layout.addWidget(self._no_type_hint, 8, 0)

        form = QFormLayout()
        group_box_layout.addLayout(form, 0, 1)

        self._py_name = QLineEdit()
        form.addRow('PyName', self._py_name)

        self._pyqt_no_qmetaobject = QCheckBox('PyQtNoQMetaObject')
        group_box_layout.addWidget(self._pyqt_no_qmetaobject, 1, 1)

        form = QFormLayout()
        group_box_layout.addLayout(form, 2, 1, 8, 1)

        self._pyqt_interface = QLineEdit()
        form.addRow('PyQtInterface', self._pyqt_interface)

        self._supertype = QLineEdit()
        form.addRow('Supertype', self._supertype)

        self._type_hint = QLineEdit()
        form.addRow('TypeHint', self._type_hint)

        self._type_hint_in = QLineEdit()
        form.addRow('TypeHintIn', self._type_hint_in)

        self._type_hint_out = QLineEdit()
        form.addRow('TypeHintOut', self._type_hint_out)

        self._type_hint_value = QLineEdit()
        form.addRow('TypeHintValue', self._type_hint_value)

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        self._py_bases.setText(self.model.pybases)

        for name, value in split_annos(self.model.annos):
            if name == 'Abstract':
                self._abstract.setChecked(True)
            elif name == 'AllowNone':
                self._allow_none.setChecked(True)
            elif name == 'DelayDtor':
                self._delay_dtor.setChecked(True)
            elif name == 'NoDefaultCtors':
                self._no_default_ctors.setChecked(True)
            elif name == 'NoTypeHint':
                self._no_type_hint.setChecked(True)
            elif name == 'PyQtNoQMetaObject':
                self._pyqt_no_qmetaobject.setChecked(True)
            elif name == 'ExportDerived':
                self._export_derived.setChecked(True)
            elif name == 'Mixin':
                self._mixin.setChecked(True)
            elif name == 'PyName':
                self._py_name.setText(Value)
            elif name == 'API':
                self._api.setText(Value)
            elif name == 'Metatype':
                self._metatype.setText(Value)
            elif name == 'Supertype':
                self._supertype.setText(Value)
            elif name == 'PyQtInterface':
                self._pyqt_interface.setText(Value)
            elif name == 'TypeHint':
                self._type_hint.setText(Value)
            elif name == 'TypeHintIn':
                self._type_hint_in.setText(Value)
            elif name == 'TypeHintOut':
                self._type_hint_out.setText(Value)
            elif name == 'TypeHintValue':
                self._type_hint_value.setText(Value)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        self.model.pybases = ' '.join(self._py_bases.text().strip().split())

        annos_list = []

        if self._abstract.isChecked():
            annos_list.append('Abstract')

        if self._allow_none.isChecked():
            annos_list.append('AllowNone')

        if self._delay_dtor.isChecked():
            annos_list.append('DelayDtor')

        if self._no_default_ctors.isChecked():
            annos_list.append('NoDefaultCtors')

        if self._no_type_hint.isChecked():
            annos_list.append('NoTypeHint')

        if self._pyqt_no_qmetaobject.isChecked():
            annos_list.append('PyQtNoQMetaObject')

        if self._export_derived.isChecked():
            annos_list.append('ExportDerived')

        if self._mixin.isChecked():
            annos_list.append('Mixin')

        py_name = self._py_name.text().strip()
        if py_name:
            annos_list.append('PyName=' + py_name)

        api = self._api.text().strip()
        if api:
            annos_list.append('API=' + api)

        metatype = self._metatype.text().strip()
        if metatype:
            annos_list.append('Metatype=' + metatype)

        supertype = self._supertype.text().strip()
        if supertype:
            annos_list.append('Supertype=' + supertype)

        type_hint = self._type_hint.text().strip()
        if type_hint:
            annos_list.append(f'TypeHint="{type_hint}"')

        type_hint_in = self._type_hint_in.text().strip()
        if type_hint_in:
            annos_list.append(f'TypeHintIn="{type_hint_in}"')

        type_hint_out = self._type_hint_out.text().strip()
        if type_hint_out:
            annos_list.append(f'TypeHintOut="{type_hint_out}"')

        type_hint_value = self._type_hint_value.text().strip()
        if type_hint_value:
            annos_list.append(f'TypeHintValue="{type_hint_value}"')

        pyqt_interface = self._pyqt_interface.text().strip()
        if pyqt_interface:
            annos_list.append('PyQtInterface=' + pyqt_interface)

        self.model.annos = ','.join(annos_list)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>



from ...shell import EventType
from ...shell_tool import ShellTool, ShellToolLocation

from .gui import ScannerGui


class ScannerTool(ShellTool):
    """ This class implements a tool for handling the scanning of a project's
    header directories.
    """

    def __init__(self, shell):
        """ Initialise the tool. """

        super().__init__(shell)

        self._gui = ScannerGui(self)

    def event(self, event_type, event_arg):
        """ Reimplemented to handle project-specific events. """

        if event_type in (EventType.MODULE_ADD, EventType.MODULE_DELETE):
            self._gui.control_widget.module_add_delete()
        elif event_type is EventType.MODULE_RENAME:
            self._gui.control_widget.module_rename(event_arg)
        elif event_type is EventType.PROJECT_NEW:
            self._set_project()
        elif event_type in (EventType.VERSION_ADD, EventType.VERSION_DELETE):
            self._gui.control_widget.version_add_delete()
        elif event_type is EventType.VERSION_RENAME:
            self._gui.control_widget.version_rename(*event_arg)

    def header_directory_added(self, header_directory, working_version):
        """ A header directory has been added. """

        self._gui.sources_widget.header_directory_added(header_directory,
                working_version)

    def header_directories_status(self, header_file):
        """ The status of a header directory has changed. """

        self._gui.sources_widget.header_directory_status()

    def header_directory_removed(self, header_directory):
        """ A header directory has been removed. """

        self._gui.sources_widget.header_directory_removed(header_directory)

    def header_file_added(self, header_file, header_directory, working_version):
        """ A header file has been added. """

        self._gui.sources_widget.header_file_added(header_file,
                header_directory, working_version)

    def header_file_status(self, header_file):
        """ The status of a header file has changed. """

        self._gui.sources_widget.header_file_status(header_file)

    def header_file_removed(self, header_file):
        """ A header file has been removed. """

        self._gui.sources_widget.header_directory_removed(header_file)

    @property
    def location(self):
        """ Get the location of the tool in the shell. """

        return ShellToolLocation.RIGHT

    @property
    def name(self):
        """ Get the tool's name. """

        return 'metasip.tool.scanner'

    def restore_state(self, settings):
        """ Restore the tool's state from the settings. """

        self._gui.restore_state(settings)

    def save_state(self, settings):
        """ Save the tool's state in the settings. """

        self._gui.save_state(settings)

    def set_header_file(self, header_file, header_directory, showing_ignored):
        """ Set the current header file. """

        self._gui.control_widget.set_header_file(header_file, header_directory,
                showing_ignored)

    def set_header_files_visibility(self, header_directory, showing_ignored):
        """ Show or hide all the ignored files in a header directory. """

        self._gui.sources_widget.set_header_files_visibility(header_directory,
                showing_ignored)

    def set_working_version(self, working_version):
        """ Set the current working version. """

        self._gui.sources_widget.set_working_version(working_version)
        self._gui.control_widget.set_working_version(working_version)

    @property
    def title(self):
        """ Get the tool's title. """

        return "Scanner"

    @property
    def widget(self):
        """ Get the tool's widget. """

        return self._gui

    def _set_project(self):
        """ Set the current project. """

        self._gui.sources_widget.set_project()
        self._gui.control_widget.set_project()

        project = self.shell.project

        # Default to the latest version if any.
        working_version = project.versions[-1] if len(project.versions) != 0 else None
        self.set_working_version(working_version)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QComboBox, QFormLayout, QLineEdit

from ...helpers import BaseDialog
from ...shell import EventType

from .helpers import init_version_selector, validate_version_name


class NewVersionDialog(BaseDialog):
    """ This class implements the dialog for creating a new version. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        form = QFormLayout()
        layout.addLayout(form)

        self._version = QLineEdit()
        form.addRow("Version name", self._version)

        self._after = QComboBox()
        form.addRow("Add version after", self._after)

    def set_fields(self):
        """ Set the dialog's fields from the project. """

        project = self.model

        init_version_selector(self._after, self.model)

        if len(project.versions) != 0:
            self._after.setEnabled(True)
            self._after.setCurrentText(project.versions[-1])
        else:
            self._after.setEnabled(False)

    def get_fields(self):
        """ Update the project from the dialog's fields. """

        project = self.model

        version = self._version.text().strip()
        if not validate_version_name(version, project, self):
            return False

        # Add to the header file versions.
        if self._after.isEnabled():
            after = self._after.currentText()

            # Create a new header file version based on the previous one.
            for hdir in project.headers:
                scan = False

                for hfile in hdir.content:
                    for hfile_version in hfile.versions:
                        if hfile_version == after:
                            new_hfile_version = HeaderFileVersion(
                                    md5=hfile_version.md5, parse=False,
                                    version=version)
                            hfile.versions.append(new_hfile_version)

                            # The header directory needs scanning.
                            scan = True

                            break

                if scan:
                    hdir.scan.append(version)

            project.versions.insert(project.versions.index(after) + 1, version)
        else:
            # The first explicit version has been defined.
            for hdir in project.headers:
                for hfile in hdir.content:
                    # There will only be one version at most defined.
                    if len(hfile.versions) != 0:
                        hfile.versions[0].version = version

            project.versions.append(version)

        self.shell.notify(EventType.VERSION_ADD, version)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from .argument_properties_dialog import ArgumentPropertiesDialog
from .callable_properties_dialog import CallablePropertiesDialog
from .class_properties_dialog import ClassPropertiesDialog
from .enum_properties_dialog import EnumPropertiesDialog
from .enum_member_properties_dialog import EnumMemberPropertiesDialog
from .features_dialog import FeaturesDialog
from .manual_code_dialog import ManualCodeDialog
from .module_properties_dialog import ModulePropertiesDialog
from .move_header_dialog import MoveHeaderDialog
from .namespace_properties_dialog import NamespacePropertiesDialog
from .opaque_class_properties_dialog import OpaqueClassPropertiesDialog
from .platforms_dialog import PlatformsDialog
from .project_properties_dialog import ProjectPropertiesDialog
from .typedef_properties_dialog import TypedefPropertiesDialog
from .variable_properties_dialog import VariablePropertiesDialog
from .versions_dialog import VersionsDialog



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass, field

from .header_file import HeaderFile
from .platform import Platform


@dataclass
class HeaderDirectory:
    """ This class implements a directory containing C/C++ .h files. """

    # The list of C/C++ .h files in the directory.
    content: list[HeaderFile] = field(default_factory=list)

    # The platform-specific configurations.
    platforms: list[Platform] = field(default_factory=list)

    # The name of the header directory.  This is used for display purposes.
    name: str = ''

    # The versions for which the header directory needs scanning.  A single
    # empty string means that no explicit versions have been defined but the
    # header directory needs scanning.
    scan: list[str] = field(default_factory=list)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


import os

from ..exceptions import UserException
from ..helpers import VersionMap, version_range
from ..models import Enum, Function, Variable
from ..models.adapters import adapt

from .indent_file import IndentFile


def generate_sip_files(project, output_dir, ignored_modules, verbose):
    """ Generate the .sip files for a project. """

    if ignored_modules is None:
        ignored_modules = []

    for module in project.modules:
        # See if the module should be ignored.
        if module.name in ignored_modules:
            if verbose:
                print(f"Ignoring {module.name}")

            continue

        # Create the module-specific output directory.
        if project.version >= (0, 17):
            module_output_dir = os.path.join(output_dir, module.name)
        else:
            if module.outputdirsuffix != '':
                module_output_dir = os.path.join(output_dir, module.outputdirsuffix)
            else:
                module_output_dir = output_dir

        # Make sure the output directory exists.
        try:
            os.makedirs(module_output_dir)
        except:
            pass

        # Generate .sip files for the module contents.
        sip_file_names = []

        for sip_file in module.content:
            (file_name, _) = os.path.splitext(os.path.basename(sip_file.name))
            file_name += '.sip'
            sip_file_names.append(file_name)

            output = _create_sip_file(project, module, module_output_dir,
                    file_name, verbose)
            _generate_sip(sip_file, project, output)
            output.close()

        # Generate the .sip file defining the module itself.
        output = _create_sip_file(project, module, module_output_dir,
                module.name + 'mod.sip', verbose)

        root_name = project.rootmodule

        if root_name != '':
            root_name += "."

        output.write('%Module(name=' + root_name + module.name)

        if module.callsuperinit != 'undefined':
            output.write(', call_super_init=' + ('True' if module.callsuperinit == 'yes' else 'False'))

        if module.virtualerrorhandler != '':
            output.write(', default_VirtualErrorHandler=' + module.virtualerrorhandler)

        if module.keywordarguments != '':
            output.write(f', keyword_arguments="{module.keywordarguments}"')

        if module.uselimitedapi:
            output.write(', use_limited_api=True')

        if module.pyssizetclean:
            output.write(', py_ssize_t_clean=True')

        output.write(')\n\n')

        top_level_module = True

        if module.imports:
            for imported in module.imports:
                output.write(f'%Import {imported}/{imported}mod.sip\n')

                if imported not in project.externalmodules:
                    top_level_module = False

            output.write('\n')

        if top_level_module:
            # Add any version, platform and feature information to all top
            # level modules (ie. those that don't import anything).

            if project.versions:
                versions = ' '.join(project.versions)
                output.write(f'%Timeline {{{versions}}}\n\n')

            if project.platforms:
                platforms = ' '.join(project.platforms)
                output.write(f'%Platforms {{{platforms}}}\n\n')

            if project.features:
                for feature in project.features:
                    output.write(f'%Feature {feature}\n')

                output.write('\n')

        if module.directives != '':
            output.write(module.directives)
            output.write('\n\n')

        for file_name in sip_file_names:
            output.write(f'%Include {file_name}\n')

        output.close()


def _create_sip_file(project, module, module_output_dir, file_name, verbose):
    """ Create and return a boilerplate .sip file. """

    if verbose:
        print(f"Generating '{file_name}'")

    output = _IndentSipFile.create(os.path.join(module_output_dir, file_name))

    # Add the standard header.
    output.write(
f'''// {file_name} generated by MetaSIP
//
// This file is part of the {module.name} Python extension module.
''')

    if project.sipcomments:
        output.write(f'//\n{project.sipcomments}\n')

    output.write('\n')
    output.blank()

    return output


def _generate_sip(sip_file, project, output):
    """ Generate the contents of a .sip file. """

    # See if we need a %ModuleCode directive for things which will be
    # implemented at the module level.  At the same time find the version
    # ranges that cover all the API items.
    vmap = VersionMap(project)
    platforms = set()
    features = set()
    need_header = False

    for api in sip_file.content:
        if api.status != '':
            continue

        # Note that OperatorFunctions are handled within the class even if they
        # have global declarations.
        if isinstance(api, (Enum, Function, Variable)):
            need_header = True

        vmap.update_from_version_ranges(api.versions)

        if platforms is not None:
            if api.platforms:
                platforms.update(api.platforms)
            else:
                platforms = None

        if features is not None:
            if api.features:
                features.update(api.features)
            else:
                features = None

    if need_header:
        vranges_str = [version_range(vr) for vr in vmap.as_version_ranges()]

        plat_feat = []

        if platforms is not None:
            plat_feat.extend(platforms)

        if features is not None:
            plat_feat.extend(features)

        for vr_str in vranges_str:
            output.write(f'%If ({vr_str})\n', indent=False)

        if plat_feat:
            tags = ' || '.join(plat_feat)
            output.write(f'%If ({tags})\n', indent=False)

        output.write(
f'''%ModuleCode
#include <{sip_file.name}>
%End
''')

        if plat_feat:
            output.write('%End\n', indent=False)

        for _ in vranges_str:
            output.write('%End\n', indent=False)

        output.blank()

    for api in sip_file.content:
        if api.status == '':
            adapt(api).generate_sip(sip_file, output)

    output.blank()

    output.write_code_directive('%ExportedHeaderCode',
            sip_file.exportedheadercode, indent=False)
    output.write_code_directive('%ModuleHeaderCode', sip_file.moduleheadercode,
            indent=False)
    output.write_code_directive('%ModuleCode', sip_file.modulecode,
            indent=False)
    output.write_code_directive('%PreInitialisationCode', sip_file.preinitcode,
            indent=False)
    output.write_code_directive('%InitialisationCode', sip_file.initcode,
            indent=False)
    output.write_code_directive('%PostInitialisationCode',
            sip_file.postinitcode, indent=False)
    output.write_code_directive('%ExportedTypeHintCode',
            sip_file.exportedtypehintcode, indent=False)
    output.write_code_directive('%TypeHintCode', sip_file.typehintcode,
            indent=False)


class _IndentSipFile(IndentFile):
    """ An indentation file with extra functionality for writing .sip files.
    """

    def write_code_directive(self, directive, code, indent=True):
        """ Write a code directive. """

        if code != '':
            self.write(directive + '\n', indent=False)
            self += 1
            self.write(code + '\n', indent=indent)
            self -= 1
            self.write('%End\n', indent=False)
            self.blank()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .access import Access
from .code import Code
from .code_container import CodeContainer
from .docstring import Docstring


@dataclass
class Class(Code, CodeContainer, Docstring, Access):
    """ This class implements a class. """

    # The C++ base classes.
    bases: str = ''

    # The optional %BIGetCharBufferCode (deprecated).
    bicharbufcode: str = ''

    # The optional %BIGetBufferCode.
    bigetbufcode: str = ''

    # The optional %BIReadBufferCode (deprecated).
    bireadbufcode: str = ''

    # The optional %BIReleaseBufferCode.
    birelbufcode: str = ''

    # The optional %BIGetSegCountCode (deprecated).
    bisegcountcode: str = ''

    # The optional %BIWriteBufferCode (deprecated).
    biwritebufcode: str = ''

    # The optional %ConvertFromTypeCode.
    convfromtypecode: str = ''

    # The optional %ConvertToTypeCode.
    convtotypecode: str = ''

    # The optional %GCClearCode.
    gcclearcode: str = ''

    # The optional %GCTraverseCode.
    gctraversecode: str = ''

    # The name of the class.
    name: str = ''

    # The optional %PickleCode.
    picklecode: str = ''

    # The optional Python base classes.  An empty string means that the C++
    # base classes should be used.  A value of "None" means that there are no
    # base classes.
    pybases: str = ''

    # This is set if the class is a struct.
    struct: bool = False

    # The optional %ConvertToSubClassCode.
    subclasscode: str = ''

    # The optional %FinalisationCode.
    finalisationcode: str = ''

    # The optional %TypeCode.
    typecode: str = ''

    # The optional %TypeHeaderCode.
    typeheadercode: str = ''

    # The optional %TypeHintCode.
    typehintcode: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import (QCheckBox, QComboBox, QFormLayout, QGroupBox,
        QLineEdit, QVBoxLayout)

from ....helpers import BaseDialog

from .helpers import BaseType, split_annos


class EnumPropertiesDialog(BaseDialog):
    """ This class implements the dialog for an enum's properties. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        group_box = QGroupBox("Annotations")
        group_box_layout = QVBoxLayout()
        group_box.setLayout(group_box_layout)
        layout.addWidget(group_box)

        form = QFormLayout()
        group_box_layout.addLayout(form)

        self._py_name = QLineEdit()
        form.addRow('PyName', self._py_name)

        base_type = QComboBox()
        form.addRow('BaseType', base_type)
        self._base_type_helper = BaseType(base_type)

        self._no_scope = QCheckBox('NoScope')
        group_box_layout.addWidget(self._no_scope)

        self._no_type_hint = QCheckBox('NoTypeHint')
        group_box_layout.addWidget(self._no_type_hint)

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        for name, value in split_annos(self.model.annos):
            if name == 'PyName':
                self._py_name.setText(value)
            elif name == 'BaseType':
                self._base_type_helper.setAnnotation(value)
            elif name == 'NoScope':
                self._no_scope.setChecked(True)
            elif name == 'NoTypeHint':
                self._no_type_hint.setChecked(True)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        annos_list = []

        self._base_type_helper.annotation(annos_list)

        s = self._py_name.text().strip()
        if s != '':
            annos_list.append('PyName=' + s)

        if self._no_scope.isChecked():
            annos_list.append('NoScope')

        if self._no_type_hint.isChecked():
            annos_list.append('NoTypeHint')

        self.model.annos = ','.join(annos_list)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


import os
import subprocess
import sys
import tempfile

from ....helpers import header_directory_platform
from ....project import (Function, Argument, Variable, Typedef, OpaqueClass,
        Class, Constructor, Destructor, Method, Enum, EnumValue,
        OperatorFunction, OperatorMethod, Namespace, OperatorCast)

from .parser_base import ParserBase, optAttribute


# The tuple of code items generated by Qt's Q_OBJECT macro to automatically
# ignore.
_Q_OBJECT = (
    # These are for Qt v3 only.
    "virtual QMetaObject *metaObject() const",
    "virtual const char *className() const",
    "virtual void *qt_cast(const char *)",
    "virtual bool qt_invoke(int, QUObject *)",
    "virtual bool qt_emit(int, QUObject *)",
    "virtual bool qt_property(int, int, QVariant *)",
    "static bool qt_static_property(QObject *, int, int, QVariant *)",
    "static QMetaObject *staticMetaObject()",
    "QObject *qObject()",

    # These are for Qt v3 and v4.
    "static QString tr(const char *, const char * = 0)",
    "static QString trUtf8(const char *, const char * = 0)",

    # These are for Qt v4 only.
    "virtual const QMetaObject *metaObject() const",
    "static const QMetaObject staticMetaObject",
    "virtual void *qt_metacast(const char *)",
    "virtual int qt_metacall(QMetaObject::Call, int, void **)",

    # These are for Qt v5.5 and later.  We don't include the argument as its
    # type varies.
    "const char *qt_getEnumName",
    "const QMetaObject *qt_getEnumMetaObject",
)


def _fixQt(code):
    """ Update the status for code items generated from Qt's Q_OBJECT macro.
    At the moment we don't check argument lists as they haven't been set up
    yet.

    :param code:
        is the code instance.
    """

    def leading(s):
        """ Return the start of the string up to the first opening parenthesis.
        """

        idx = s.find("(")

        if idx > 0:
            s = s[:idx]

        return s

    cs = leading(code.user())

    for qo in _Q_OBJECT:
        if cs == leading(qo):
            code.status = "ignored"
            break


class _Access(object):
    """
    This class is derived by all code that is affected by class access.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the instance.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        self.access = attrs.get('access', '')
        if self.access == 'public':
            self.access = ''


class _ScopedItem(object):
    """
    This class is a base class for any entity that is part of a scope.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the instance.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        # Not everything has a name (some structs for example).
        try:
            self.name = attrs["name"]
        except KeyError:
            self.name = None

        # Not everything has an ID.
        try:
            self.id = attrs["id"]
        except KeyError:
            self.id = None

        if self.id:
            parser.byid[self.id] = self

        # The root namespace doesn't have a context.
        try:
            self.context = attrs["context"]
        except KeyError:
            self.context = None

        # Namespaces don't have file IDs or line numbers.
        try:
            self.file = attrs["file"]
        except KeyError:
            self.file = None

        try:
            self.line = int(attrs["line"])
        except KeyError:
            self.line = None

        parser.scopeditems.append(self)

    def asType(self, parser, prefix_ok):
        """
        Return the string representation of the item for when it is used as a
        type.

        parser is the parser instance.
        """
        assert prefix_ok is None

        sl = []

        # Prefix it by any scope.
        if self.name:
            sl.append(self.name)

        pc = parser.byid[self.context]

        # Watch for the root namespace which doesn't have a context.
        while isinstance(pc, _ScopedItem) and pc.context is not None:
            sl.insert(0, pc.name)
            pc = parser.byid[pc.context]

        return "::".join(sl), True


class _Namespace(_ScopedItem):
    """
    This class represents a Cast-XML namespace entity.
    """
    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        tci = Namespace(name=self.name, container=scope)
        scope.content.append(tci)

        parser.transformScope(tci, self)


class _Class(_ScopedItem, _Access):
    """
    This class represents a Cast-XML class entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _ScopedItem.__init__(self, parser, attrs)
        _Access.__init__(self, parser, attrs)

        self.bases = optAttribute(attrs, "bases")
        self.incomplete = bool(int(optAttribute(attrs, "incomplete", "0")))

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        if self.name is None:
            return

        if self.incomplete:
            scope.content.append(
                    OpaqueClass(name=self.name, container=scope,
                            access=self.access, status='ignored'))
        else:
            bl = []

            for bid in self.bases.split():
                sbid = bid.split(":")

                if len(sbid) == 1:
                    acc = "public"
                else:
                    acc = sbid[0]
                    bid = sbid[1]

                bl.append("%s %s" % (acc, parser.asType(bid)))

            # Automatically ignore non-public classes.
            status = 'unknown' if self.access == '' else 'ignored'

            tci = Class(name=self.name, container=scope, bases=', '.join(bl),
                    struct=False, access=self.access, status=status)
            scope.content.append(tci)

            parser.transformScope(tci, self)


class _Struct(_ScopedItem, _Access):
    """
    This class represents a Cast-XML struct entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _ScopedItem.__init__(self, parser, attrs)
        _Access.__init__(self, parser, attrs)

        self.incomplete = bool(int(optAttribute(attrs, "incomplete", "0")))

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        if self.name is None:
            return

        if self.incomplete:
            scope.content.append(
                    OpaqueClass(name=self.name, container=scope,
                            access=self.access, status='ignored'))
        else:
            # Automatically ignore non-public classes.
            status = 'unknown' if self.access == '' else 'ignored'

            tci = Class(name=self.name, container=scope, struct=True,
                    access=self.access, status=status)
            scope.content.append(tci)

            parser.transformScope(tci, self)


# Treat a union as a struct so that it appears in the .sip file where it can be
# handled with handwritten code.
_Union = _Struct


class _Callable(_ScopedItem):
    """
    This class is the base class for callable code.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _ScopedItem.__init__(self, parser, attrs)

        self.args = []


class _ClassCallable(_Callable, _Access):
    """
    This class is the base class for callable code in a class context.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _Callable.__init__(self, parser, attrs)
        _Access.__init__(self, parser, attrs)


class _Constructor(_ClassCallable):
    """
    This class represents a Cast-XML constructor entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _ClassCallable.__init__(self, parser, attrs)

        self.explicit = bool(int(optAttribute(attrs, "explicit", "0")))

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        # Don't specify explicit unless it's applicable.
        if len(self.args) == 1:
            explicit = self.explicit
        else:
            explicit = False

        tci = Constructor(name=self.name, container=scope, access=self.access,
                explicit=explicit)

        _transformArgs(parser, self.args, tci.args)

        scope.content.append(tci)


class _Destructor(_ScopedItem, _Access):
    """
    This class represents a Cast-XML destructor entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _ScopedItem.__init__(self, parser, attrs)
        _Access.__init__(self, parser, attrs)

        self.virtual = bool(int(optAttribute(attrs, "virtual", "0")))

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        scope.content.append(
                Destructor(name=self.name, container=scope, access=self.access,
                        virtual=self.virtual))


class _Converter(_ClassCallable):
    """
    This class represents a Cast-XML converter entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        super().__init__(parser, attrs)

        self.returns = attrs["returns"]
        self.const = bool(int(optAttribute(attrs, "const", "0")))

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        name = parser.asType(self.returns)
        if name is not None:
            tci = OperatorCast(name=name, container=scope, access=self.access,
                    const=self.const)

            scope.content.append(tci)


class _Method(_ClassCallable):
    """
    This class represents a Cast-XML method entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _ClassCallable.__init__(self, parser, attrs)

        self.returns = attrs["returns"]
        self.virtual = bool(int(optAttribute(attrs, "virtual", "0")))
        self.const = bool(int(optAttribute(attrs, "const", "0")))
        self.static = bool(int(optAttribute(attrs, "static", "0")))
        self.abstract = bool(int(optAttribute(attrs, "pure_virtual", "0")))

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        # Private methods are ignored by default.
        status = 'ignored' if self.access.startswith('private') else 'unknown'

        tci = Method(name=self.name, container=scope, access=self.access,
                rtype=parser.asType(self.returns), virtual=self.virtual,
                const=self.const, static=self.static, abstract=self.abstract,
                status=status)

        _transformArgs(parser, self.args, tci.args)

        _fixQt(tci)

        scope.content.append(tci)


class _OperatorMethod(_ClassCallable):
    """
    This class represents a Cast-XML operatormethod entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _ClassCallable.__init__(self, parser, attrs)

        self.returns = attrs["returns"]
        self.virtual = bool(int(optAttribute(attrs, "virtual", "0")))
        self.const = bool(int(optAttribute(attrs, "const", "0")))
        self.abstract = bool(int(optAttribute(attrs, "pure_virtual", "0")))

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        # Private methods are ignored by default.
        status = 'ignored' if self.access.startswith('private') else 'unknown'

        tci = OperatorMethod(name=self.name, container=scope,
                access=self.access, rtype=parser.asType(self.returns),
                virtual=self.virtual, const=self.const, abstract=self.abstract,
                status=status)

        _transformArgs(parser, self.args, tci.args)

        scope.content.append(tci)


class _Function(_Callable):
    """
    This class represents a Cast-XML function entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _Callable.__init__(self, parser, attrs)

        self.returns = attrs["returns"]

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        tci = Function(name=self.name, container=scope,
                rtype=parser.asType(self.returns))

        _transformArgs(parser, self.args, tci.args)

        _fixQt(tci)

        scope.content.append(tci)


class _OperatorFunction(_Callable):
    """
    This class represents a Cast-XML operatorfunction entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _Callable.__init__(self, parser, attrs)

        self.returns = attrs["returns"]

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        tci = OperatorFunction(name=self.name, container=scope,
                rtype=parser.asType(self.returns))

        _transformArgs(parser, self.args, tci.args)

        scope.content.append(tci)


class _Variable(_ScopedItem, _Access):
    """
    This class represents a Cast-XML variable entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _ScopedItem.__init__(self, parser, attrs)
        _Access.__init__(self, parser, attrs)

        self.type_id = attrs["type"]

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        status = 'ignored' if self.access.startswith('protected') else 'unknown'

        tci = Variable(name=self.name, type=parser.asType(self.type_id),
                static=isinstance(scope, Class), access=self.access,
                status=status)

        _fixQt(tci)

        scope.content.append(tci)


class _Field(_Variable):
    """
    This class represents a Cast-XML field entity.
    """
    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        t = parser.asType(self.type_id)

        if t:
            scope.content.append(
                    Variable(name=self.name, type=t, static=False,
                            access=self.access))


class _Enumeration(_ScopedItem, _Access):
    """
    This class represents a Cast-XML enumeration entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _ScopedItem.__init__(self, parser, attrs)
        _Access.__init__(self, parser, attrs)

        # Deal with anonymous enums.
        if self.name.startswith("."):
            self.name = ""

        self.scoped = bool(int(attrs.get('scoped', '0')))
        self.values = []

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        # We don't support private enums.  We didn't filter it out before
        # because it may have been used as a type.
        if self.access.startswith("private"):
            return

        tci = Enum(name=self.name, access=self.access, enumclass=self.scoped)

        for e in self.values:
            tci.content.append(EnumValue(name=e.name))

        scope.content.append(tci)


class _EnumValue(object):
    """
    This class represents a Cast-XML enum value entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        self.name = attrs["name"]


class _Typedef(_ScopedItem):
    """
    This class represents a Cast-XML typedef entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _ScopedItem.__init__(self, parser, attrs)

        self.type_id = attrs["type"]

    def transform(self, parser, scope):
        """
        Transform the entity.

        parser is the parser instance.
        scope is the scope to append the transformed entity to.
        """
        t = parser.asType(self.type_id)

        # Ignore unsupported types - probably only those defined in terms of
        # a MethodType (eg. typedef foo_t (scope::*bar)();).
        if t:
            scope.content.append(Typedef(name=self.name, type=t))

    def asType(self, parser, prefix_ok):
        """
        Return the string representation of the item for when it is used as a
        type.

        parser is the parser instance.
        """
        # enum_type (as a reference to the underlying QFlags<> argument) was
        # introduced in Qt v5.6.  We simply replace it with the enum to make
        # sure signatures don't change from previous Qt versions.
        if self.name == 'enum_type':
            return parser.byid[self.type_id].asType(parser, prefix_ok)

        return _ScopedItem.asType(self, parser, prefix_ok)


class _FunctionType(object):
    """
    This class represents a Cast-XML function type entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        self.returns = attrs["returns"]
        self.args = []

        parser.byid[attrs["id"]] = self

    def asType(self, parser, prefix_ok):
        """
        Return the string representation of the type.

        parser is the parser instance.
        """
        assert prefix_ok is None

        al = []
        for a in self.args:
            al.append(parser.asType(a.type_id))

        return parser.asType(self.returns) + " (%s)(" + ", ".join(al) + ")", False


class _FundamentalType(object):
    """
    This class represents a Cast-XML fundamental type entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        self.name = attrs["name"]

        parser.byid[attrs["id"]] = self

    def asType(self, parser, prefix_ok):
        """
        Return the string representation of the type.

        parser is the parser instance.
        """
        assert prefix_ok is None

        # Map some of Cast-XML's verbose types to something SIP can handle.
        type_map = {
            "short int": "short",
            "short unsigned int": "unsigned short",
            "long unsigned int": "unsigned long",
            "long long unsigned int": "unsigned long long"
        }

        return type_map.get(self.name, self.name), True


class _IndirectType(object):
    """
    This class represents the base type for all Cast-XML indirect type
    entities, ie. those linked to other types through a type attribute.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        self.type_id = attrs["type"]

        parser.byid[attrs["id"]] = self

    def asType(self, parser, prefix_ok):
        """
        Return the string representation of the type that this type indirects
        to.

        parser is the parser instance.
        """
        return parser.asInnerType(self.type_id, prefix_ok)


class _ReferenceType(_IndirectType):
    """
    This class represents a Cast-XML reference type entity.
    """
    def asType(self, parser, prefix_ok):
        """
        Return the string representation of the type.

        parser is the parser instance.
        """
        s, _ = super().asType(parser, prefix_ok)

        if s[-1] not in "*&":
            s += " "

        return s + "&", False


class _PointerType(_IndirectType):
    """
    This class represents a Cast-XML pointer type entity.
    """
    def asType(self, parser, prefix_ok):
        """
        Return the string representation of the type.

        parser is the parser instance.
        """
        s, _ = super().asType(parser, prefix_ok)

        rt = parser.byid[self.type_id]

        if isinstance(rt, _FunctionType):
            # If the base type is a function then put the pointer immediately
            # before the function name marker.
            s = s % "*%s"
        else:
            if s[-1] not in "*&":
                s += " "

            s += "*"

        return s, False


class _CvQualifiedType(_IndirectType):
    """
    This class represents a Cast-XML reference type entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        _IndirectType.__init__(self, parser, attrs)

        self.const = bool(int(optAttribute(attrs, "const", "0")))

    def asType(self, parser, prefix_ok):
        """
        Return the string representation of the type.

        parser is the parser instance.
        """
        s, prefix_ok = super().asType(parser, prefix_ok)

        if self.const:
            if prefix_ok:
                s = "const " + s
            else:
                s = s + " const"

        return s, False


class _Argument(object):
    """
    This class represents a Cast-XML argument entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        self.type_id = attrs["type"]
        self.name = optAttribute(attrs, "name")
        self.default = optAttribute(attrs, "default")

        # Negative numbers are represented as hex for some reason.
        if self.default.startswith('-0x'):
            self.default = str(int(self.default, base=16))

        # These are Qt specific and should be moved to a plugin.
        if self.default == 'QLatin1Char(32)':
            self.default = "QLatin1Char(' ')"

        # This is the expansion of Q_NULLPTR introduced in Qt v5.6.
        if self.default == '__null':
            self.default = '0'

        # This is the expansion of Q_NULLPTR introduced in Qt v5.7.
        if self.default == 'nullptr':
            self.default = '0'


class _Ellipsis(object):
    """
    This class represents a Cast-XML ellipsis entity.
    """
    def __init__(self, parser, attrs):
        """
        Initialise the entity.

        parser is the parser instance.
        attr is the entity's attribute dictionary.
        """
        pass


def _transformArgs(parser, gargs, pargs):
    """
    Transform a list of Cast-XML arguments and append them to a list of project
    arguments.

    parser is the parser instance.
    gargs is the list of Cast-XML arguments.
    pargs is the list of project arguments.
    """

    for a in gargs:
        if isinstance(a, _Ellipsis):
            pa = Argument(type="...")
        else:
            # Cast-XML doesn't add the scope to default values of enums so we
            # try and fix it here.
            full_type = parser.asType(a.type_id)
            if full_type is None:
                continue

            # For Qt5.
            if full_type.endswith('::QPrivateSignal'):
                continue

            pa = Argument(type=full_type, name=a.name, default=a.default)

        pargs.append(pa)


class _CodeContainer:
    """ An internal class that implements the root of the transformed items.
    """

    def __init__(self, project):
        """ Initialise the object. """

        self.project = project
        self.content = []


class CastXMLParser(ParserBase):
    """
    This class implements a C++ parser based on Cast-XML.  It should be used as
    an abstract class with a derived class handling any user interaction.
    """
    def classMap(self):
        """
        Return a dictionary of XML entity names and their corresponding
        classes.  These are the ones that don't need any special handling.
        """
        return {"class":            _Class,
                "typedef":          _Typedef,
                "fundamentaltype":  _FundamentalType,
                "referencetype":    _ReferenceType,
                "pointertype":      _PointerType,
                "arraytype":        _PointerType,
                "cvqualifiedtype":  _CvQualifiedType}

    def parse(self, project, input_dir, hdir, hf, pathname, log):
        """
        Parse a file and return the parsed file instance or None if there was
        an error.

        project is the project.
        input_dir is the root input directory.
        hdir is the header directory instance.
        hf is the header file instance.
        pathname is the name of the actual file to parse.
        """
        self._pathname = pathname
        iname = os.path.join(tempfile.gettempdir(), hf.name + '.tmp')

        argv = ['castxml', '-x', 'c++', '--castxml-output=1']

        if sys.platform == 'darwin':
            # Workaround issues with Xcode v14.3 (and possibly later).
            argv.append('-Dat_quick_exit=atexit')
            argv.append('-Dquick_exit=exit')

            xcode = subprocess.check_output(('xcode-select', '-p')).decode().strip()

            argv.append('-isysroot')
            argv.append(xcode + '/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk')

        argv.append('-o')
        argv.append(iname)
        argv.append(header_directory_platform(hdir).parserargs)
        argv.append(self._pathname)

        # We use shell=True and a string argv for macOS - but I don't
        # understand why it's needed.
        args = ' '.join(argv)

        log(args)

        cwd = os.getcwd()
        os.chdir(input_dir)

        try:
            output = subprocess.check_output(args, shell=True,
                    stderr=subprocess.STDOUT)
            rc = 0
        except subprocess.CalledProcessError as exc:
            output = exc.output
            rc = exc.returncode

        os.chdir(cwd)

        # Log any output.
        for line in output.decode().rstrip().split('\n'):
            log(line.rstrip())

        if rc != 0:
            try:
                os.remove(iname)
            except:
                pass

            sig = rc & 0x7f
            rc >>= 8

            if sig:
                self.diagnostic = "{0} killed by signal {1}".format(argv[0],
                        sig)
            else:
                self.diagnostic = "{0} failed with exit code {1}".format(
                        argv[0], rc)

            log(self.diagnostic)

            return None

        log("Parsing XML for {0}".format(self._pathname))

        # Initialise the parser state.  The first pass is to read in the
        # Cast-XML output filtering out stuff we definately don't need.  The
        # second pass is to convert it into the internal project format.  We
        # need to do two passes because Cast-XML does not ensure that
        # everything is defined by the time it is referenced.
        self.byid = {}
        self.scopeditems = []
        self._rootns = None
        self._args = None
        self._evalues = None
        self._fileid = None

        rc = super().parse(iname)

        #os.remove(iname)

        if not rc:
            return None

        # Now convert it to the internal format.
        phf = _CodeContainer(project)

        self.transformScope(phf, self._rootns)

        return phf.content

    def namespaceStart(self, attrs):
        """
        Called at the start of a namespace.

        attrs is the dictionary of attributes.
        """
        ns = _Namespace(self, attrs)

        # Remember the root namespace.
        if ns.name == '::':
            self._rootns = ns

    def fieldStart(self, attrs):
        """
        Called at the start of a field.

        attrs is the dictionary of attributes.
        """
        # At the moment we only support public variables.  We could support
        # protected ones in the future, but we will never support private ones.
        if optAttribute(attrs, "access") == "private":
            return

        _Field(self, attrs)

    def variableStart(self, attrs):
        """
        Called at the start of a variable.

        attrs is the dictionary of attributes.
        """
        # At the moment we only support public variables.  We could support
        # protected ones in the future, but we will never support private ones.
        if optAttribute(attrs, "access") == "private":
            return

        if optAttribute(attrs, "artificial", None) is not None:
            return

        _Variable(self, attrs)

    def unionStart(self, attrs):
        """
        Called at the start of a union.

        attrs is the dictionary of attributes.
        """
        _Union(self, attrs)

    def structStart(self, attrs):
        """
        Called at the start of a structure.

        attrs is the dictionary of attributes.
        """
        _Struct(self, attrs)

    def constructorStart(self, attrs):
        """
        Called at the start of a constructor.

        attrs is the dictionary of attributes.
        """
        if optAttribute(attrs, "artificial", None) is not None:
            return

        self._args = _Constructor(self, attrs).args

    def constructorEnd(self):
        """
        Called at the end of a constructor.
        """
        self._args = None

    def destructorStart(self, attrs):
        """
        Called at the start of a destructor.

        attrs is the dictionary of attributes.
        """
        if optAttribute(attrs, "artificial", None) is not None:
            return

        _Destructor(self, attrs)

    def converterStart(self, attrs):
        """
        Called at the start of a converter.

        attrs is the dictionary of attributes.
        """
        _Converter(self, attrs)

    def methodStart(self, attrs):
        """
        Called at the start of a method.

        attrs is the dictionary of attributes.
        """
        self._args = _Method(self, attrs).args

    def methodEnd(self):
        """
        Called at the end of a method.
        """
        self._args = None

    def operatormethodStart(self, attrs):
        """
        Called at the start of an operatormethod.

        attrs is the dictionary of attributes.
        """
        self._args = _OperatorMethod(self, attrs).args

    def operatormethodEnd(self):
        """
        Called at the end of an operatormethod.
        """
        self._args = None

    def functionStart(self, attrs):
        """
        Called at the start of a function.

        attrs is the dictionary of attributes.
        """
        self._args = _Function(self, attrs).args

    def functionEnd(self):
        """
        Called at the end of a function.
        """
        self._args = None

    def functiontypeStart(self, attrs):
        """
        Called at the start of a function type.

        attrs is the dictionary of attributes.
        """
        self._args = _FunctionType(self, attrs).args

    def functiontypeEnd(self):
        """
        Called at the end of a function type.
        """
        self._args = None

    def operatorfunctionStart(self, attrs):
        """
        Called at the start of an operatorfunction.

        attrs is the dictionary of attributes.
        """
        self._args = _OperatorFunction(self, attrs).args

    def operatorfunctionEnd(self):
        """
        Called at the end of an operatorfunction.
        """
        self._args = None

    def argumentStart(self, attrs):
        """
        Called at the start of an argument.

        attrs is the dictionary of attributes.
        """
        if self._args is not None:
            self._args.append(_Argument(self, attrs))

    def ellipsisStart(self, attrs):
        """
        Called at the start of an ellipsis.

        attrs is the dictionary of attributes.
        """
        if self._args is not None:
            self._args.append(_Ellipsis(self, attrs))

    def enumerationStart(self, attrs):
        """
        Called at the start of an enumeration.

        attrs is the dictionary of attributes.
        """
        self._evalues = _Enumeration(self, attrs).values

    def enumvalueStart(self, attrs):
        """
        Called at the start of an enum value.

        attrs is the dictionary of attributes.
        """
        self._evalues.append(_EnumValue(self, attrs))

    def fileStart(self, attrs):
        """
        Called at the start of a file.

        attrs is the dictionary of attributes.
        """
        # Remember the file ID of the file being parsed.
        name = attrs["name"]
        if os.path.isabs(name):
            if name == self._pathname:
                self._fileid = attrs["id"]
        else:
            if self._pathname.endswith(name[1:]):
                self._fileid = attrs["id"]

    def transformScope(self, container, scope):
        """
        Transform a scope (either a namespace or a class) from the stripped
        down Cast-XML format to the internal project format.

        container is the code container to add to.
        scope is the scope to convert.
        """
        # Transform each scoped item.
        for si in self._sorted_scope(scope):
            si.transform(self, container)

    def _sorted_scope(self, unsorted):
        """
        Return a list of the items in a scope sorted by line number.

        unsorted is the scope.
        """
        # Return the cached list if we have already built it.
        try:
            return unsorted._cache
        except AttributeError:
            pass

        ssl = []

        for si in self.scopeditems:
            # Skip if this item is part of the scope.
            if si is unsorted or si.context != unsorted.id:
                continue

            # If we don't know the item's position then it must be a namespace
            # so sort it's contents and take the position of it's first
            # "child".
            if si.file is None or si.line is None:
                si_sorted = self._sorted_scope(si)

                # Skip it if it was empty.
                if len(si_sorted) == 0:
                    continue

                si.file = si_sorted[0].file
                si.line = si_sorted[0].line

            if si.file == self._fileid:
                ssl.append(si)

        ssl.sort(key=lambda k: k.line)
        unsorted._cache = ssl

        return ssl

    def asType(self, type_id):
        """
        Return the string representation of a type.

        type_id is the type ID.
        """
        type_str, _ = self.asInnerType(type_id, None)

        return type_str

    def asInnerType(self, type_id, prefix_ok):
        """
        Return the string representation of a type.

        type_id is the type ID.
        """
        try:
            type_str, prefix_ok = self.byid[type_id].asType(self, prefix_ok)
        except KeyError:
            type_str = 'unknown_type'

        return type_str, prefix_ok



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


import os

from PyQt6.QtGui import QAction
from PyQt6.QtWidgets import QFileDialog

from ....models import Project
from ....project_io import load_project, save_project

from ...helpers import ProjectUi
from ...shell import EventType
from ...shell_tool import ShellTool, ShellToolLocation

from .api_editor import ApiEditor


class ApiEditorTool(ShellTool):
    """ This class implements the API editor tool. """

    def __init__(self, shell):
        """ Initialise the tool. """

        super().__init__(shell)

        self._api_editor = ApiEditor(shell)

    @property
    def actions(self):
        """ Get the destination menu and sequence of actions handled by the
        tool.
        """

        new_action = QAction("New", self._api_editor,
                triggered=self._handle_new)
        open_action = QAction("Open...", self._api_editor,
                triggered=self._handle_open)
        save_action = QAction("Save", self._api_editor,
                triggered=self._handle_save)
        save_as_action = QAction("Save as...", self._api_editor,
                triggered=self._handle_save_as)

        return "File", (new_action, open_action, save_action, save_as_action)

    def event(self, event_type, event_arg):
        """ Reimplemented to handle project-specific events. """

        if event_type is EventType.API_STATUS:
            self._api_editor.api_status(event_arg)
        elif event_type is EventType.API_VERSIONS:
            self._api_editor.api_versions(event_arg)
        elif event_type is EventType.CONTAINER_API_ADD:
            self._api_editor.container_api_add(*event_arg)
        elif event_type is EventType.CONTAINER_API_DELETE:
            self._api_editor.container_api_delete(*event_arg)
        elif event_type is EventType.MODULE_ADD:
            self._api_editor.module_add(event_arg)
        elif event_type is EventType.MODULE_DELETE:
            self._api_editor.module_delete(event_arg)
        elif event_type is EventType.MODULE_RENAME:
            self._api_editor.module_rename(event_arg)
        elif event_type is EventType.PROJECT_NEW:
            self._api_editor.set_project()
        elif event_type is EventType.PROJECT_ROOT_MODULE_RENAME:
            self._api_editor.root_module_updated()

    @property
    def location(self):
        """ Get the location of the tool in the shell. """

        return ShellToolLocation.CENTRE

    @property
    def name(self):
        """ Get the tool's name. """

        return 'metasip.tool.api_editor'

    def restore_state(self, settings):
        """ Restore the tool's state from the settings. """

        self._api_editor.restore_state(settings)

    def save_data(self):
        """ Save the tool's data and return True if there was no error. """

        saved = save_project(self.shell.project, ui=ProjectUi())
        if saved:
            self.shell.dirty = False

        return saved

    def save_state(self, settings):
        """ Save the tool's state in the settings. """

        self._api_editor.save_state(settings)

    @property
    def title(self):
        """ Get the tool's title. """

        return "API Editor"

    @property
    def widget(self):
        """ Get the tool's widget. """

        return self._api_editor

    def _handle_new(self):
        """ Handle the New action. """

        if self.shell.save_project():
            self.shell.project = Project()

    def _handle_open(self):
        """ Handle the Open action. """

        if self.shell.save_project():
            dir_name = os.path.dirname(self.shell.project.name)

            project_name, _ = QFileDialog.getOpenFileName(self._api_editor,
                    "Open project file", dir_name,
                    "MetaSIP project files (*.msp)")

            if project_name:
                project = Project(name=project_name)
                if load_project(project, ui=ProjectUi()):
                    self.shell.project = project

    def _handle_save(self):
        """ Handle the Save action. """

        if self.shell.project.name != '':
            self.save_data()
        else:
            self._handle_save_as()

    def _handle_save_as(self):
        """ Handle the Save as action. """

        dir_name = os.path.dirname(self.shell.project.name)

        project_name, _ = QFileDialog.getSaveFileName(self._api_editor,
                "Save project file", dir_name, "MetaSIP project files (*.msp)")

        if project_name:
            self.shell.project.name = project_name
            self.shell.notify(EventType.PROJECT_RENAME)
            self.save_data()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..header_directory import HeaderDirectory
from ..module import Module
from ..project_version import ProjectVersion

from .adapt import adapt
from .base_adapter import AttributeType, BaseAdapter


class ProjectAdapter(BaseAdapter):
    """ This is the Project adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'externalfeatures':     AttributeType.STRING_LIST,
        'externalmodules':      AttributeType.STRING_LIST,
        'features':             AttributeType.STRING_LIST,
        'platforms':            AttributeType.STRING_LIST,
        'rootmodule':           AttributeType.STRING,
        'sipcomments':          AttributeType.LITERAL,
        'versions':             AttributeType.STRING_LIST,
    }

    def as_str(self):
        """ Return the standard string representation. """

        return self.model.rootmodule

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        # Initialise any UI for the load.
        if ui is not None:
            # Each .sip file is a step of the load.
            ui.load_starting(self.model, len(element.findall('.//SipFile')))

        super().load(element, project, ui)

        for subelement in element:
            if subelement.tag == 'HeaderDirectory':
                header_directory = HeaderDirectory()
                adapt(header_directory).load(subelement, project, ui)
                self.model.headers.append(header_directory)
            elif subelement.tag == 'Module':
                module = Module()
                adapt(module).load(subelement, project, ui)
                self.model.modules.append(module)

    def save(self, output):
        """ Save the model to an output file. """

        project = self.model

        # Note that we always use the current project version.
        major_version, minor_version = ProjectVersion
        if major_version == 0:
            format_version = f'version="{minor_version}"'
        else:
            format_version = f'majorversion="{major_version}" minorversion="{minor_version}"'

        output.write('<?xml version="1.0"?>\n')
        output.write(
                f'<Project {format_version} rootmodule="{project.rootmodule}"')
        self.save_str_list('versions', output)
        self.save_str_list('platforms', output)
        self.save_str_list('features', output)
        self.save_str_list('externalmodules', output)
        self.save_str_list('externalfeatures', output)
        output.write('>\n')
        output += 1

        self.save_literal('sipcomments', output)

        for header_directory in project.headers:
            adapt(header_directory).save(output)

        for module in project.modules:
            adapt(module).save(output)

        output -= 1
        output.write('</Project>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QComboBox, QDialog, QFormLayout

from ....helpers import BaseDialog


class MoveHeaderDialog(BaseDialog):
    """ This class implements the dialog for moving a header file to another
    module.
    """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        project = self.shell.project

        form = QFormLayout()
        layout.addLayout(form)

        self._dst_module = QComboBox()
        form.addRow("Destination module", self._dst_module)

        for module in sorted(project.modules, key=lambda m: m.name):
            if module is not self.model:
                self._dst_module.addItem(module.name, module)

    def get_destination_module(self):
        """ Return the destination module or None if the dialog was cancelled.
        """

        if self.widget.exec() == int(QDialog.DialogCode.Rejected):
            return None

        return self._dst_module.currentData()



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from xml.sax import make_parser, SAXParseException
from xml.sax.handler import ContentHandler, ErrorHandler


class ParserBase(ContentHandler):
    """
    The base class for all MetaSIP XML parsers.  Sub-classes implement methods
    with the same name as the elements they deal with (with "Start" or "End"
    appended).
    """
    def __init__(self):
        """
        Initialise the parser instance.
        """
        ContentHandler.__init__(self)

        self.diagnostic = None

        self._parser = make_parser()
        self._parser.setContentHandler(self)

    def classMap(self):
        """
        The default implementation to return an empty class map.
        """
        return {}

    def _findClass(self, name):
        """
        Return the class corresponding to the given name, or None if none.

        name is the name.
        """
        return self.classMap().get(name)

    def _findMethod(self, name):
        """
        Return the bound method with the given name.
        """
        try:
            m = getattr(self, name)
        except AttributeError:
            m = None

        return m

    def startElement(self, name, attrs):
        """
        The start-of-an-element callback.

        name is the element name.
        attrs is the dictionary of attributes.
        """
        name = name.lower()

        m = self._findClass(name)

        if m:
            m(self, attrs)
        else:
            m = self._findMethod(name + "Start")

            if m:
                m(attrs)

    def endElement(self, name):
        """
        The end-of-an-element callback.

        name is the element name.
        """
        name = name.lower()

        if self._findClass(name):
            return

        m = self._findMethod(name + "End")

        if m:
            m()

    def parse(self, ifname):
        """ Parse an XML file.

        :param ifname:
            is the name of a file or a file object to parse.
        :return:
            ``True`` if the file was parsed successfully.
        """

        eh = _ParserErrorHandler()

        self._parser.setErrorHandler(eh)

        try:
            self._parser.parse(ifname)
        except SAXParseException:
            self.diagnostic = eh.diagnostic
            return False

        return True


class _ParserErrorHandler(ErrorHandler):
    """ This is the SAX error handler. """

    def __init__(self):
        """ Initialise the error handler. """

        self.diagnostic = None

    def error(self, exception):
        """ Handle recoverable errors (where parsing might continue). """

        xs = str(exception)

        self.shell.log(xs)

        if not self.diagnostic:
            self.diagnostic = xs

        raise exception

    def fatalError(self, exception):
        """ Handle non-recoverable errors (where parsing can't continue). """

        xs = str(exception)

        self.shell.log(xs)

        if not self.diagnostic:
            self.diagnostic = xs

        raise exception

    def warning(self, exception):
        """ Handle warnings. """

        self.shell.log(str(exception))


def optAttribute(attrs, name, default=''):
    """ Return the value of an optional attribute.

    :param attrs:
        is the dictionary of attributes.
    :param name:
        is the name of the attribute to return.
    :param default:
        is the value to return if the attribute wasn't specified.
    :return:
        the value of the attribute.
    """

    try:
        return attrs[name]
    except KeyError:
        return default



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QAbstractItemView, QTreeWidget, QTreeWidgetItem


class SourcesWidget(QTreeWidget):
    """ This class is a widget that implements a browser of the header files of
    a source directory.
    """

    # The column numbers.
    (NAME, STATUS) = range(2)

    def __init__(self, tool):
        """ Initialise the widget. """

        super().__init__()

        self._tool = tool

        self.setHeaderLabels(("Name", "Status"))

        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.currentItemChanged.connect(self._handle_selection_change)

    def header_directory_added(self, header_directory, working_version):
        """ A header directory has been added. """

        header_directory_item = _HeaderDirectoryItem(header_directory, self)
        header_directory_item.set_working_version(working_version)
        self._sort_header_directories()

    def header_directories_status(self, header_directory):
        """ The status of a header directory has changed. """

        header_directory_item = self._find_header_directory_item(
                header_directory)
        header_directory_item.set_status()

    def header_directory_removed(self, header_directory):
        """ A header directory has been removed. """

        header_directory_item = self._find_header_directory_item(
                header_directory)
        header_directory_index = self.indexOfTopLevelItem(
                header_directory_item)
        self.takeTopLevelItem(header_directory_index)

    def header_file_added(self, header_file, header_directory, working_version):
        """ A header file has been added. """

        header_directory_item = self._find_header_directory_item(
                header_directory)
        header_file_item = _HeaderFileItem(header_file, header_directory_item)
        header_file_item.set_working_version(working_version)
        header_directory_item.sort_header_files()

    def header_file_status(self, header_file):
        """ The status of a header file has changed. """

        header_file_item = self._find_header_file_item(header_file)
        header_file_item.set_status()

    def header_file_removed(self, header_file):
        """ A header file has been removed. """

        header_file_item = self._find_header_file_item(header_file)

        if len(header_file.versions) == 0:
            # The header files doesn't exist in any version.
            header_directory_item = header_file_item.parent()
            header_file_index = header_directory_item.indexOfChild(
                    header_file_item)
            header_directory_item.takeChild(header_item_index)
        else:
            # The header file still exists in other versions.
            header_file_item.setHidden(True)

    def restore_state(self, settings):
        """ Restore the widget's state. """

        state = settings.value('header')
        if state is not None:
            self.header().restoreState(state)

    def save_state(self, settings):
        """ Save the widget's state. """

        settings.setValue('header', self.header().saveState())

    def set_header_files_visibility(self, header_directory, showing_ignored):
        """ Show or hide all the ignored files in a header directory. """

        header_directory_item = self._find_header_directory_item(
                header_directory)

        header_directory_item.showing_ignored = showing_ignored

        for header_file_item in self._header_file_items(header_directory_item):
            if header_file_item.project_item.ignored:
                header_file_item.setHidden(not showing_ignored)

    def set_project(self):
        """ Set the current project. """

        self.clear()

        for header_directory in self._tool.shell.project.headers:
            _HeaderDirectoryItem(header_directory, self)

        self._sort_header_directories()

    def set_working_version(self, working_version):
        """ Set the current working version. """

        for header_directory_item in self._header_directory_items():
            header_directory_item.set_working_version(working_version)

            for header_file_item in self._header_file_items(header_directory_item):
                header_file_item.set_working_version(working_version)

    def _find_header_directory_item(self, header_directory):
        """ Return the header directory item for a header directory. """

        for header_directory_item in self._header_directory_items():
            if header_directory_item.project_item is header_directory:
                return header_directory_item

        # This should never happen.
        return None

    def _find_header_file_item(self, header_file):
        """ Return the header file item for a header file. """

        for header_directory_item in self._header_directory_items():
            for header_file_item in self._header_file_items(header_directory_item):
                if header_file_item.project_item is header_file:
                    return header_file_item

        # This should never happen.
        return None

    def _handle_selection_change(self, current, previous):
        """ Invoked when the item selection changes. """

        project_item = current.project_item

        if self.indexOfTopLevelItem(current) >= 0:
            header_directory_item = current
            header_file = None
        else:
            header_directory_item = current.parent()
            header_file = project_item

        self._tool.set_header_file(header_file,
                header_directory_item.project_item,
                header_directory_item.showing_ignored)

    def _header_directory_items(self):
        """ A generator for the header directory items. """

        for idx in range(self.topLevelItemCount()):
            yield self.topLevelItem(idx)

    def _header_file_items(self, header_directory_item):
        """ A generator for the header file items of a header directory. """

        for idx in range(header_directory_item.childCount()):
            yield header_directory_item.child(idx)

    def _sort_header_directories(self):
        """ Sort the header directoroes in order. """

        self.sortItems(SourcesWidget.NAME, Qt.SortOrder.AscendingOrder)


class _SourcesItem(QTreeWidgetItem):
    """ This is a base class for all items in the sources tree. """

    def __init__(self, project_item, parent):
        """ Initialise the item. """

        super().__init__(parent)

        self.project_item = project_item
        self.working_version = None

    def set_status(self):
        """ Set the status of the item. """

        raise NotImplementedError

    def set_working_version(self, working_version):
        """ Set the current working versions. """

        self.working_version = working_version
        self.set_status()


class _HeaderDirectoryItem(_SourcesItem):
    """ This class represents a header directory in the sources tree. """

    def __init__(self, header_directory, parent):
        """ Initialise the item. """

        super().__init__(header_directory, parent)

        self.showing_ignored = False

        self.setText(SourcesWidget.NAME, header_directory.name)

        for header_file in header_directory.content:
            _HeaderFileItem(header_file, self)

        self.sort_header_files()

    def set_status(self):
        """ Set the status of the item. """

        self.setExpanded(False)

        # Draw the status column.
        if self.working_version is None:
            needs_scanning = (len(self.project_item.scan) != 0)
        else:
            needs_scanning = (self.working_version in self.project_item.scan)

        self.setText(SourcesWidget.STATUS,
                "Needs scanning" if needs_scanning else '')

    def sort_header_files(self):
        """ Sort the header files in the directory. """

        self.sortChildren(SourcesWidget.NAME, Qt.SortOrder.AscendingOrder)


class _HeaderFileItem(_SourcesItem):
    """ This class represents a header file in the sources tree. """

    def __init__(self, header_file, parent):
        """ Initialise the item. """

        super().__init__(header_file, parent)

        self.setText(SourcesWidget.NAME, header_file.name)

    def set_status(self):
        """ Set the status of the item. """

        hidden = False
        expand_parent = False
        status = ''

        if self.project_item.ignored:
            status = "Ignored"
            hidden = not self.parent().showing_ignored
        elif self.project_item.module == '':
            status = "Needs assigning"
            expand_parent = True
        else:
            working_header_file = self._get_working_header_file()

            if working_header_file is None:
                hidden = True
            elif working_header_file.parse:
                status = "Needs parsing"
                expand_parent = True

        self.setText(SourcesWidget.STATUS, status)
        self.setHidden(hidden)

        if expand_parent:
            self.parent().setExpanded(True)

    def _get_working_header_file(self):
        """ Return the version header file for the current working version if
        any.
        """

        header_file_versions = self.project_item.versions

        if self.working_version is None:
            working_file = None if self.project_item.ignored else header_file_versions[0]
        else:
            for working_file in header_file_versions:
                if working_file.version == self.working_version:
                    break
            else:
                working_file = None

        return working_file



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtGui import QAction

from ...shell import EventType
from ...shell_tool import ActionTool


class ModulesTool(ActionTool):
    """ This class implements the modules tool. """

    @property
    def actions(self):
        """ Get the destination menu and sequence of actions handled by the
        tool.
        """

        self._new_action = QAction("New Module...",
                triggered=self._handle_new)
        self._rename_action = QAction("Rename Module...",
                triggered=self._handle_rename)
        self._delete_action = QAction("Delete Module...",
                triggered=self._handle_delete)

        return ("Edit",
                (self._new_action, self._rename_action, self._delete_action))

    def event(self, event_type, event_arg):
        """ Reimplemented to handle project-specific events. """

        if event_type in (EventType.PROJECT_NEW, EventType.MODULE_ADD, EventType.MODULE_DELETE):
            # Configure the actions.
            are_modules = (len(self.shell.project.modules) != 0)
            self._rename_action.setEnabled(are_modules)
            self._delete_action.setEnabled(are_modules)

    @property
    def name(self):
        """ Get the tool's name. """

        return 'metasip.tool.modules'

    def _handle_delete(self):
        """ Handle the Delete action. """

        from .delete_module_dialog import DeleteModuleDialog

        self.shell.handle_project_dialog("Delete Module", DeleteModuleDialog)

    def _handle_new(self):
        """ Handle the New action. """

        from .new_module_dialog import NewModuleDialog

        self.shell.handle_project_dialog("New Module", NewModuleDialog)

    def _handle_rename(self):
        """ Handle the Rename action. """

        from .rename_module_dialog import RenameModuleDialog

        self.shell.handle_project_dialog("Rename Module", RenameModuleDialog)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


import argparse
import sys

try:
    from PyQt6.QtWidgets import QApplication
except ImportError:
    print("Run \"pip install 'metasip[gui]'\" to install the additional GUI packages", file=sys.stderr)
    sys.exit(1)

from ..exceptions import UserException
from ..models import Project
from ..project_io import load_project

from .helpers import ProjectUi, warning
from .shell import Shell
from .tools import (ApiEditorTool, FeaturesTool, ImportProjectTool, LoggerTool,
        ModulesTool, PlatformsTool, ScannerTool, VersionsTool)


def main():
    """ The entry point for the msip gui script. """

    # Parse the command line.
    parser = argparse.ArgumentParser()
    parser.add_argument('project', help="the project to edit", nargs='?')
    args = parser.parse_args()

    project_name = args.project

    app = QApplication(sys.argv, applicationName='metasip',
            organizationDomain='riverbankcomputing.com',
            organizationName='Riverbank Computing')

    sys.excepthook = _exception_hook

    # Load any project.
    project = Project(project_name)
    if not load_project(project, ui=ProjectUi()):
        return 0

    # Create the shell.
    shell = Shell(ApiEditorTool, FeaturesTool, ImportProjectTool, LoggerTool,
            ModulesTool, PlatformsTool, ScannerTool, VersionsTool)

    # Set the project in the shell.
    shell.project = project

    # Launch the GUI.
    shell.show()
    return app.exec()


def _exception_hook(exc_type, exc_value, exc_tb):
    """ Handle an exception. """

    if isinstance(exc_value, UserException):
        from .utils import warning

        warning('metasip', exc_value.text, detail=exc_value.detail)
    else:
        import traceback

        tb = ''.join(traceback.format_exception(exc_type, exc_value, exc_tb))

        print(tb, file=sys.stderr)

        QApplication.exit(1)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass, field

from .sip_file import SipFile


@dataclass
class Module:
    """ This class implements a Python module. """

    # Whether instances of classes defined in the module should call
    # super().__init().  Values are 'undefined', 'no' and 'yes'.
    callsuperinit: str = ''

    # The list of .sip files defining API items included in the module.
    content: list[SipFile] = field(default_factory=list)

    # The SIP directives to be included at the start of the main .sip file for
    # the module.
    directives: str = ''

    # The list of modules that this module depends on.
    imports: list[str] = field(default_factory=list)

    # The default handling of keyword arguments, either '', 'all' or
    # 'optional'.
    # Added in v0.17
    keywordarguments: str = ''

    # The name of the module.
    name: str = ''

    # The optional suffix added to the output directory to create the full name
    # of the directory where the generated .sip files will be placed.
    # Removed in v0.17
    outputdirsuffix: str = ''

    # Set if the module is PY_SSIZE_T_CLEAN.
    pyssizetclean: bool = False

    # Set if the limited Python API should be used.
    uselimitedapi: bool = False

    # The default virtual error handler.
    virtualerrorhandler: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .code import Code
from .code_container import CodeContainer


@dataclass
class Namespace(Code, CodeContainer):
    """ This class implements a namespace. """

    # The name of the namespace.
    name: str = ''

    # The optional %TypeHeaderCode.
    typeheadercode: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..helpers import validate_identifier, validation_error


def init_platform_selector(selector, project):
    """ Initialise a platform selector. """

    selector.clear()
    selector.addItems(sorted(project.platforms))


def validate_platform_name(platform, project, dialog):
    """ Validate a platform name and return True if it is valid. """

    message = validate_identifier(platform, "platform")
    if message is None:
        if platform in project.platforms:
            message = "A platform has already been defined with the same name."

    if message is not None:
        validation_error(message, dialog)
        return False

    return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from .base_adapter import AttributeType, BaseAdapter


class PlatformAdapter(BaseAdapter):
    """ This is the Platform adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'inputdirpattern':  AttributeType.STRING,
        'name':             AttributeType.STRING,
        'parserargs':       AttributeType.STRING,
    }

    def save(self, output):
        """ Save the model to an output file. """

        output.write(f'<Platform name="{self.model.name}"')
        self.save_str('inputdirpattern', output)
        self.save_str('parserargs', output)
        output.write('/>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .code import Code
from .docstring import Docstring
from .extended_access import ExtendedAccess


@dataclass
class ManualCode(Code, Docstring, ExtendedAccess):
    """ This class implements an explicitly written API. """

    # The optional explicitly written API item.  If the API item can be
    # specified in a single line then the precis attribute is normally used
    # instead.
    body: str = ''

    # The optional %MethodCode.  This would not normally be specified if the
    # body attribute is not an empty string.
    methcode: str = ''

    # The one-line summary of the code.  If the body attribute is an empty
    # string then this is copied to the .sip file unaltered.  If the body
    # attribute is not an empty string then this is copied to the .sip file as
    # a comment above the body.
    precis: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .callable import Callable
from .docstring import Docstring
from .extended_access import ExtendedAccess


@dataclass
class Method(Callable, Docstring, ExtendedAccess):
    """ This class implements a C++ class method. """

    # This is set if the method is abstract.
    abstract: bool = False

    # This is set if the method is const.
    const: bool = False

    # This is set if the method is final.
    final: bool = False

    # This is set if the method is static.
    static: bool = False

    # The optional %VirtualCatcherCode.
    virtcode: str = ''

    # This is set if the method is virtual.
    virtual: bool = False



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..exceptions import UserException


class IndentFile:
    """ This is a thin wrapper around a file object that supports indentation.
    """

    def __init__(self, file_name, indent):
        """ Create a file for writing. """

        self._f = open(file_name, 'w', encoding='UTF-8')
        self._indent = indent
        self._nr_indents = 0
        self._indent_next = True
        self._blank = False
        self._suppress_blank = False

        self.name = file_name

    def __iadd__(self, by):
        """ Increase the indentation. """

        self._nr_indents += by
        self._suppress_blank = True

        return self

    def __isub__(self, by):
        """ Decrease the indentation. """

        self._nr_indents -= by
        self._blank = False
        self._suppress_blank = False

        return self

    def blank(self):
        """ Write a blank line. """

        if not self._suppress_blank:
            self._blank = True

    def close(self):
        """ Close the file. """

        self._f.close()

    @classmethod
    def create(cls, file_name, indent=4):
        """ Return an indent file or raise an exception if there was an error.
        """

        try:
            return cls(file_name, indent)
        except IOError as e:
            raise UserException(f"There was an error creating '{file_name}'",
                    detail=str(e)) from e

    def write(self, data, indent=True):
        """ Write data to the file with optional automatic indentation. """

        if data:
            if self._blank:
                self._f.write('\n')
                self._blank = False

            lines = data.split('\n')

            for l in lines[:-1]:
                if indent and self._indent_next:
                    self._f.write(' ' * (self._indent * self._nr_indents))

                self._f.write(l + '\n')
                self._indent_next = True

            # Handle the last line.
            l = lines[-1]

            if l:
                if indent and self._indent_next:
                    self._f.write(' ' * (self._indent * self._nr_indents))

                self._f.write(l)
                self._indent_next = False
            else:
                self._indent_next = True

            self._suppress_blank = False



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


import argparse
import sys
import os

from .exceptions import UserException
from .models import Project
from .project_io import generate_sip_files, load_project
from ._version import version


def main():
    """ The entry point for the msipgen console script. """

    # Parse the command line.
    parser = argparse.ArgumentParser()

    parser.add_argument('-V', '--version', action='version', version=version)
    parser.add_argument('project',
            help="the project to generate .sip files from",
            nargs='?')
    parser.add_argument('--ignore',
            help="do not generate .sip files for MODULE",
            metavar='MODULE', action='append')
    parser.add_argument('--output-dir', help="generate the .sip files in DIR",
            metavar='DIR', required=True)
    parser.add_argument('--verbose', help="display progress messages",
            dest='verbose', default=False, action='store_true')

    args = parser.parse_args()

    try:
        _generate(args.project, args.output_dir, args.ignore, args.verbose)
    except Exception as e:
        _handle_exception(e)


def _generate(project_name, output_dir, ignore, verbose):
    """ Generate the .sip files for a project and return an exit code or 0 if
    there was no error.
    """

    if not project_name:
        raise UserException("Specify the name of an existing project file")

    project = Project(project_name)
    load_project(project)

    generate_sip_files(project, output_dir, ignore, verbose)


def _handle_exception(e):
    """ Tell the user about an exception. """

    if isinstance(e, UserException):
        # An "expected" exception.
        if e.detail is not None:
            message = "{0}: {1}".format(e.text, e.detail)
        else:
            message = e.text

        print("{0}: {1}".format(os.path.basename(sys.argv[0]), message),
                file=sys.stderr)

        sys.exit(1)

    # An internal error.
    print("{0}: An internal error occurred...".format(
            os.path.basename(sys.argv[0])),
            file=sys.stderr)

    raise e



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import (QCheckBox, QComboBox, QFormLayout, QGridLayout,
        QGroupBox, QLineEdit)

from .....models import (Constructor, Destructor, Method, OperatorMethod,
        Function, OperatorFunction, ManualCode)

from ....helpers import BaseDialog

from .helpers import Encoding, split_annos


class CallablePropertiesDialog(BaseDialog):
    """ This class implements the dialog for callable properties. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        form = QFormLayout()
        layout.addLayout(form)

        self._py_type = QLineEdit()
        form.addRow("Python return type", self._py_type)

        self._py_args = QLineEdit()
        form.addRow("Python arguments", self._py_args)

        group_box = QGroupBox("Annotations")
        layout.addWidget(group_box)
        group_box_layout = QGridLayout()
        group_box.setLayout(group_box_layout)

        self._abort_on_exception = QCheckBox('AbortOnException')
        group_box_layout.addWidget(self._abort_on_exception, 0, 0)

        self._allow_none = QCheckBox('AllowNone')
        group_box_layout.addWidget(self._allow_none, 1, 0)

        form = QFormLayout()
        group_box_layout.addLayout(form, 2, 0)

        self._api = QLineEdit()
        form.addRow('API', self._api)

        self._auto_gen = QLineEdit()
        form.addRow('AutoGen', self._auto_gen)

        self._default = QCheckBox('Default')
        group_box_layout.addWidget(self._default, 3, 0)

        self._disallow_none = QCheckBox('DisallowNone')
        group_box_layout.addWidget(self._disallow_none, 4, 0)

        form = QFormLayout()
        group_box_layout.addLayout(form, 5, 0, 2, 1)

        self._doc_type = QLineEdit()
        form.addRow('DocType', self._doc_type)

        self._encoding = QComboBox()
        form.addRow('Encoding', self._encoding)
        self._encoding_helper = Encoding(self._encoding)

        self._factory = QCheckBox('Factory')
        group_box_layout.addWidget(self._factory, 7, 0)

        # Note that 'final' isn't actually an annotation.
        self._final = QCheckBox('final')
        group_box_layout.addWidget(self._final, 8, 0)

        self._hold_gil = QCheckBox('HoldGIL')
        group_box_layout.addWidget(self._hold_gil, 9, 0)

        self._keep_reference = QCheckBox('KeepReference')
        group_box_layout.addWidget(self._keep_reference, 10, 0)

        self._new_thread = QCheckBox('NewThread')
        group_box_layout.addWidget(self._new_thread, 11, 0)

        self._no_arg_parser = QCheckBox('NoArgParser')
        group_box_layout.addWidget(self._no_arg_parser, 12, 0)

        self._no_copy = QCheckBox('NoCopy')
        group_box_layout.addWidget(self._no_copy, 13, 0)

        self._no_derived = QCheckBox('NoDerived')
        group_box_layout.addWidget(self._no_derived, 14, 0)

        self._no_type_hint = QCheckBox('NoTypeHint')
        group_box_layout.addWidget(self._no_type_hint, 0, 1)

        self._numeric = QCheckBox('Numeric')
        group_box_layout.addWidget(self._numeric, 1, 1)

        form = QFormLayout()
        group_box_layout.addLayout(form, 2, 1, 4, 1)

        self._post_hook = QLineEdit()
        form.addRow('PostHook', self._post_hook)

        self._pre_hook = QLineEdit()
        form.addRow('PreHook', self._pre_hook)

        self._py_name = QLineEdit()
        form.addRow('PyName', self._py_name)

        self._pyqt_signal_hack = QLineEdit()
        form.addRow('PyQtSignalHack', self._pyqt_signal_hack)

        self._py_int = QCheckBox('PyInt')
        group_box_layout.addWidget(self._py_int, 6, 1)

        self._release_gil = QCheckBox('ReleaseGIL')
        group_box_layout.addWidget(self._release_gil, 7, 1)

        self._transfer = QCheckBox('Transfer')
        group_box_layout.addWidget(self._transfer, 8, 1)

        self._transfer_back = QCheckBox('TransferBack')
        group_box_layout.addWidget(self._transfer_back, 9, 1)

        self._transfer_this = QCheckBox('TransferThis')
        group_box_layout.addWidget(self._transfer_this, 10, 1)

        form = QFormLayout()
        group_box_layout.addLayout(form, 11, 1)

        self._type_hint = QLineEdit()
        form.addRow('TypeHint', self._type_hint)

        self._len = QCheckBox('__len__')
        group_box_layout.addWidget(self._len, 12, 1)

        self._matmul = QCheckBox('__matmul__')
        group_box_layout.addWidget(self._matmul, 13, 1)

        self._imatmul = QCheckBox('__imatmul__')
        group_box_layout.addWidget(self._imatmul, 14, 1)

        if isinstance(self.model, Constructor):
            self.setWindowTitle("Constructor Properties")

            self._py_type.setEnabled(False)
            self._abort_on_exception.setEnabled(False)
            self._allow_none.setEnabled(False)
            self._auto_gen.setEnabled(False)
            self._disallow_none.setEnabled(False)
            self._doc_type.setEnabled(False)
            self._encoding.setEnabled(False)
            self._factory.setEnabled(False)
            self._final.setEnabled(False)
            self._keep_reference.setEnabled(False)
            self._len.setEnabled(False)
            self._matmul.setEnabled(False)
            self._imatmul.setEnabled(False)
            self._new_thread.setEnabled(False)
            self._no_arg_parser.setEnabled(False)
            self._no_copy.setEnabled(False)
            self._numeric.setEnabled(False)
            self._py_int.setEnabled(False)
            self._transfer_back.setEnabled(False)
            self._transfer_this.setEnabled(False)
            self._py_name.setEnabled(False)
            self._type_hint.setEnabled(False)

        elif isinstance(self.model, Destructor):
            self.setWindowTitle("Destructor Properties")

            self._py_type.setEnabled(False)
            self._py_args.setEnabled(False)
            self._abort_on_exception.setEnabled(False)
            self._allow_none.setEnabled(False)
            self._api.setEnabled(False)
            self._auto_gen.setEnabled(False)
            self._default.setEnabled(False)
            self._disallow_none.setEnabled(False)
            self._doc_type.setEnabled(False)
            self._encoding.setEnabled(False)
            self._factory.setEnabled(False)
            self._final.setEnabled(False)
            self._keep_reference.setEnabled(False)
            self._len.setEnabled(False)
            self._matmul.setEnabled(False)
            self._imatmul.setEnabled(False)
            self._new_thread.setEnabled(False)
            self._no_arg_parser.setEnabled(False)
            self._no_copy.setEnabled(False)
            self._no_derived.setEnabled(False)
            self._numeric.setEnabled(False)
            self._post_hook.setEnabled(False)
            self._pre_hook.setEnabled(False)
            self._py_int.setEnabled(False)
            self._transfer.setEnabled(False)
            self._transfer_back.setEnabled(False)
            self._transfer_this.setEnabled(False)
            self._py_name.setEnabled(False)
            self._type_hint.setEnabled(False)

        elif isinstance(self.model, Method):
            self.setWindowTitle("Method Properties")

            self._default.setEnabled(False)
            self._no_derived.setEnabled(False)

        elif isinstance(self.model, OperatorMethod):
            self.setWindowTitle("Operator Method Properties")

            self._abort_on_exception.setEnabled(False)
            self._default.setEnabled(False)
            self._final.setEnabled(False)
            self._keep_reference.setEnabled(False)
            self._len.setEnabled(False)
            self._matmul.setEnabled(False)
            self._imatmul.setEnabled(False)
            self._no_arg_parser.setEnabled(False)
            self._no_derived.setEnabled(False)
            self._py_name.setEnabled(False)

        elif isinstance(self.model, Function):
            self.setWindowTitle("Function Properties")

            self._abort_on_exception.setEnabled(False)
            self._default.setEnabled(False)
            self._final.setEnabled(False)
            self._keep_reference.setEnabled(False)
            self._len.setEnabled(False)
            self._matmul.setEnabled(False)
            self._imatmul.setEnabled(False)
            self._no_copy.setEnabled(False)
            self._no_derived.setEnabled(False)
            self._transfer.setEnabled(False)
            self._transfer_this.setEnabled(False)

        elif isinstance(self.model, OperatorFunction):
            self.setWindowTitle("Operator Function Properties")

            self._abort_on_exception.setEnabled(False)
            self._default.setEnabled(False)
            self._final.setEnabled(False)
            self._keep_reference.setEnabled(False)
            self._len.setEnabled(False)
            self._matmul.setEnabled(False)
            self._imatmul.setEnabled(False)
            self._no_arg_parser.setEnabled(False)
            self._no_copy.setEnabled(False)
            self._no_derived.setEnabled(False)
            self._transfer.setEnabled(False)
            self._transfer_this.setEnabled(False)
            self._py_name.setEnabled(False)

        elif isinstance(self.model, ManualCode):
            self.setWindowTitle("Manual Code Properties")

            self._py_type.setEnabled(False)
            self._py_args.setEnabled(False)
            self._py_name.setEnabled(False)

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        if isinstance(self.model, Constructor):
            self._py_args.setText(self.model.pyargs)

        elif isinstance(self.model, Method):
            self._py_type.setText(self.model.pytype)
            self._py_args.setText(self.model.pyargs)
            self._final.setChecked(self.model.final)

        elif isinstance(self.model, (OperatorMethod, Function, OperatorFunction)):
            self._py_type.setText(self.model.pytype)
            self._py_args.setText(self.model.pyargs)

        for name, value in split_annos(self.model.annos):
            if name == 'AbortOnException':
                self._abort_on_exception.setChecked(True)
            elif name == 'AllowNone':
                self._allow_none.setChecked(True)
            elif name == 'API':
                self._api.setText(value)
            elif name == 'AutoGen':
                if value is None:
                    value = 'All'

                self._auto_gen.setText(value)
            elif name == 'Default':
                self._default.setChecked(True)
            elif name == 'DisallowNone':
                self._disallow_none.setChecked(True)
            elif name == 'DocType':
                self._doc_type.setText(value)
            elif name == 'Encoding':
                self._encoding_helper.setAnnotation(value)
            elif name == 'Factory':
                self._factory.setChecked(True)
            elif name == 'HoldGIL':
                self._hold_gil.setChecked(True)
            elif name == 'KeepReference':
                self._keep_reference.setChecked(True)
            elif name == '__len__':
                self._len.setChecked(True)
            elif name == '__matmul__':
                self._matmul.setChecked(True)
            elif name == '__imatmul__':
                self._imatmul.setChecked(True)
            elif name == 'NewThread':
                self._new_thread.setChecked(True)
            elif name == 'NoArgParser':
                self._no_arg_parser.setChecked(True)
            elif name == 'NoCopy':
                self._no_copy.setChecked(True)
            elif name == 'NoDerived':
                self._no_derived.setChecked(True)
            elif name == 'NoTypeHint':
                self.noTypeHintCb.setChecked(True)
            elif name == 'Numeric':
                self._numeric.setChecked(True)
            elif name == 'PostHook':
                self._post_hook.setText(value)
            elif name == 'PreHook':
                self._pre_hook.setText(value)
            elif name == 'PyInt':
                self._py_int.setChecked(True)
            elif name == 'PyName':
                self._py_name.setText(value)
            elif name == 'PyQtSignalHack':
                self._pyqt_signal_hack.setText(value)
            elif name == 'ReleaseGIL':
                self._release_gil.setChecked(True)
            elif name == 'Transfer':
                self._transfer.setChecked(True)
            elif name == 'TransferBack':
                self._transfer_back.setChecked(True)
            elif name == 'TransferThis':
                self._transfer_this.setChecked(True)
            elif name == 'TypeHint':
                self._type_hint.setText(value)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        if hasattr(self.model, 'pytype') and self.model.pytype is not None:
            self.model.pytype = self._py_type.text().strip()

        if hasattr(self.model, 'pyargs'):
            self.model.pyargs = self._py_args.text().strip()

        if hasattr(self.model, 'final'):
            self.model.final = self._final.isChecked()

        annos_list = []

        if self._abort_on_exception.isChecked():
            annos_list.append('AbortOnException')

        if self._allow_none.isChecked():
            annos_list.append('AllowNone')

        api = self._api.text().strip()
        if api:
            annos_list.append('API=' + api)

        auto_gen = self._auto_gen.text().strip()
        if auto_gen:
            if auto_gen == 'All':
                annos_list.append('AutoGen')
            else:
                annos_list.append('AutoGen=' + auto_gen)

        if self._default.isChecked():
            annos_list.append('Default')

        if self._disallow_none.isChecked():
            annos_list.append('DisallowNone')

        doc_type = self._doc_type.text().strip()
        if doc_type:
            annos_list.append(f'DocType="{doc_type}"')

        self._encoding_helper.annotation(annos_list)

        if self._factory.isChecked():
            annos_list.append('Factory')

        if self._hold_gil.isChecked():
            annos_list.append('HoldGIL')

        if self._keep_reference.isChecked():
            annos_list.append('KeepReference')

        if self._len.isChecked():
            annos_list.append('__len__')

        if self._matmul.isChecked():
            annos_list.append('__matmul__')

        if self._imatmul.isChecked():
            annos_list.append('__imatmul__')

        if self._new_thread.isChecked():
            annos_list.append('NewThread')

        if self._no_arg_parser.isChecked():
            annos_list.append('NoArgParser')

        if self._no_copy.isChecked():
            annos_list.append('NoCopy')

        if self._no_derived.isChecked():
            annos_list.append('NoDerived')

        if self.noTypeHintCb.isChecked():
            annos_list.append('NoTypeHint')

        if self._numeric.isChecked():
            annos_list.append('Numeric')

        post_hook = self._post_hook.text().strip()
        if post_hook:
            annos_list.append('PostHook=' + post_hook)

        pre_hook = self._pre_hook.text().strip()
        if pre_hook:
            annos_list.append('PreHook=' + pre_hook)

        if self._py_int.isChecked():
            annos_list.append('PyInt')

        py_name = self._py_name.text().strip()
        if py_name:
            annos_list.append('PyName=' + py_name)

        pyqt_signal_hack = self._pyqt_signal_hack.text().strip()
        if pyqt_signal_hack:
            annos_list.append('PyQtSignalHack=' + pyqt_signal_hack)

        if self._release_gil.isChecked():
            annos_list.append('ReleaseGIL')

        if self._transfer.isChecked():
            annos_list.append('Transfer')

        if self._transfer_back.isChecked():
            annos_list.append('TransferBack')

        if self._transfer_this.isChecked():
            annos_list.append('TransferThis')

        type_hint = self._type_hint.text().strip()
        if type_hint:
            annos_list.append(f'TypeHint="{type_hint}"')

        self.model.annos = ','.join(annos_list)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QCheckBox, QFormLayout, QLineEdit

from ....project import Module

from ...helpers import BaseDialog
from ...shell import EventType

from .helpers import validate_module_name


class NewModuleDialog(BaseDialog):
    """ This class implements the dialog for creating a new module. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        form = QFormLayout()
        layout.addLayout(form)

        self._module = QLineEdit()
        form.addRow("Module name", self._module)

        self._external = QCheckBox("Module is defined in another project?")
        layout.addWidget(self._external)

    def get_fields(self):
        """ Update the project from the dialog's fields. """

        project = self.model

        module_name = self._module.text().strip()
        if not validate_module_name(module_name, project, self):
            return False

        if self._external.isChecked():
            project.externalmodules.append(module_name)
        else:
            module = Module(name=module_name)
            project.modules.append(module)
            self.shell.notify(EventType.MODULE_ADD, module)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ...shell import EventType

from ..helpers import tagged_items, validate_identifier, validation_error


def init_version_selector(selector, project):
    """ Initialise a version selector. """

    selector.clear()
    selector.addItems(project.versions)


def validate_version_name(version, project, dialog):
    """ Validate a version name and return True if it is valid. """

    message = validate_identifier(version, "version")
    if message is None:
        if version in project.versions:
            message = "A version has already been defined with the same name."

    if message is not None:
        validation_error(message, dialog)
        return False

    return True


def delete_version(version, shell, *, migrate_items):
    """ Delete a version and all API items defined by it. """

    project = shell.project
    versions = project.versions

    removing_first_version = (version == versions[0])

    # Delete from each API item it appears.
    remove_items = []

    for api_item, container in tagged_items(project):
        # Ignore items that aren't tagged with a version.
        if len(api_item.versions) == 0:
            continue

        remove_ranges = []

        for r in api_item.versions:
            if r.startversion == version:
                # It now starts with the version after the one we are deleting.
                # If that is the same as the end version then we delete the API
                # item.
                new_start = '' if versions[-1] == version else versions[versions.index(version) + 1]

                if new_start == r.endversion:
                    remove_items.append((api_item, container))
                    break

                r.startversion = new_start

            # If the start version is now what the first version will now be
            # then we clear it.
            if r.startversion != '' and removing_first_version and r.startversion == versions[1]:
                r.startversion = ''

            if r.endversion == version:
                if migrate_items:
                    # It now ends with the version after the one we are
                    # deleting.
                    new_end = '' if versions[-1] == version else versions[versions.index(version) + 1]

                    r.endversion = new_end
                else:
                    remove_items.append((api_item, container))
                    break

            if r.startversion == '' and r.endversion == '':
                remove_ranges.append(r)
        else:
            for r in remove_ranges:
                api_item.versions.remove(r)

    for api_item, container_item in remove_items:
        container_item.content.remove(api_item)
        shell.notify(EventType.CONTAINER_API_DELETE,
                (container_item, api_item))

    # Delete from the header file versions.
    remove_hfile_versions = []
    removing_last_version = (len(project.versions) == 1)

    for hdir in project.headers:
        if version in hdir.scan:
            hdir.scan.remove(version)

        # If the versions to scan now just has a marker from when no versions
        # were defined, remove the marker so it doesn't suggest that any
        # remaining version needs scanning.
        if len(hdir.scan) == 1 and hdir.scan[0] == '':
            del hdir.scan[0]

        for hfile in hdir.content:
            if removing_last_version:
                if len(hfile.versions) != 0:
                    hfile.versions[0].version = ''
            else:
                for hfile_version in hfile.versions:
                    if hfile_version.version == version:
                        remove_hfile_versions.append((hfile_version, hfile))
                        break

    for hfile_version, hfile in remove_hfile_versions:
        hfile.versions.remove(hfile_version)

    # Delete from the project's list.
    versions.remove(version)
    shell.notify(EventType.VERSION_DELETE, version)



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import (QCheckBox, QComboBox, QFormLayout, QGridLayout,
        QGroupBox, QLineEdit, QPlainTextEdit)

from ....helpers import BaseDialog


class ModulePropertiesDialog(BaseDialog):
    """ This class implements the dialog for a module's properties. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        project = self.shell.project

        self._all_imports = list(project.externalmodules)
        self._all_imports.extend([m.name for m in project.modules])
        self._all_imports.sort()

        form = QFormLayout()
        layout.addLayout(form)

        if project.version < (0, 17):
            self._output_dir_suffix = QLineEdit()
            form.addRow("Output directory suffix", self._output_dir_suffix)

        self._directives = QPlainTextEdit()
        self._directives.setLineWrapMode(QPlainTextEdit.LineWrapMode.NoWrap)
        form.addRow("Additional directives", self._directives)

        self._call_super_init = QComboBox()
        self._call_super_init.addItems(("Undefined", "No", "Yes"))
        form.addRow("Call super().__init__()", self._call_super_init)

        self._keyword_arguments = QComboBox()
        self._keyword_arguments.addItems(("None", "All", "Optional"))
        form.addRow("Default handling of keyword arguments",
                self._keyword_arguments)

        self._virtual_error_handler = QLineEdit()
        form.addRow("Default virtual error handler",
                self._virtual_error_handler)

        self._py_ssizet_clean = QCheckBox("#define PY_SSIZE_T_CLEAN")
        layout.addWidget(self._py_ssizet_clean)

        self._use_limited_api = QCheckBox("Use the limited Python API")
        layout.addWidget(self._use_limited_api)

        group_box = QGroupBox("%Import")
        layout.addWidget(group_box)
        self._imports_layout = QGridLayout()
        group_box.setLayout(self._imports_layout)

        for i, module in enumerate(self._all_imports):
            self._imports_layout.addWidget(QCheckBox(module), i // 2, i % 2)

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        module = self.model

        if self.shell.project.version < (0, 17):
            self._output_dir_suffix.setText(module.outputdirsuffix)

        self._directives.setPlainText(module.directives)

        if module.callsuperinit == 'undefined':
            idx = 0
        elif module.callsuperinit == 'no':
            idx = 1
        else:
            idx = 2

        self._call_super_init.setCurrentIndex(idx)

        if module.keywordarguments == 'Optional':
            idx = 2
        elif module.keywordarguments == 'All':
            idx = 1
        else:
            idx = 0

        self._keyword_arguments.setCurrentIndex(idx)

        self._virtual_error_handler.setText(module.virtualerrorhandler)
        self._use_limited_api.setChecked(module.uselimitedapi)
        self._py_ssizet_clean.setChecked(module.pyssizetclean)

        for i in range(self._imports_layout.count()):
            check_box = self._imports_layout.itemAt(i).widget()
            module_name = check_box.text()

            check_box.setEnabled(module_name != module.name)
            check_box.setChecked(module_name in module.imports)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        module = self.model

        if self.shell.project.version < (0, 17):
            module.outputdirsuffix = self._output_dir_suffix.text().strip()

        module.directives = self._directives.toPlainText().strip()
        module.callsuperinit = self._call_super_init.currentText().lower()
        module.virtualerrorhandler = self._virtual_error_handler.text().strip()
        module.pyssizetclean = self._py_ssizet_clean.isChecked()
        module.uselimitedapi = self._use_limited_api.isChecked()

        module.keywordarguments = self._keyword_arguments.currentText()
        if module.keywordarguments == 'None':
            module.keywordarguments = ''

        imports_list = []

        for i in range(self._imports_layout.count()):
            if self._imports_layout.itemAt(i).widget().isChecked():
                imports_list.append(self._all_imports[i])

        module.imports = imports_list

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from abc import ABC, abstractmethod
from enum import auto, Enum


class ShellToolLocation(Enum):
    """ The different possible locations for a tool in the shell. """

    CENTRE = auto()
    LEFT = auto()
    RIGHT = auto()
    TOP = auto()
    BOTTOM = auto()


class ActionTool(ABC):
    """ A base class for tools that implement optional actions. """

    def __init__(self, shell):
        """ Initialise the tool. """

        self.shell = shell

    @property
    def actions(self):
        """ Get the destination menu and sequence of actions handled by the
        tool.
        """

        # This default implementation does nothing.
        return None, ()

    def event(self, event_type, event_arg):
        """ This is called whenever a project-specific event takes place. """

        # This default implementation does nothing.
        pass

    @property
    @abstractmethod
    def name(self):
        """ Get the internal unique name of the tool. """

        ...

    def restore_state(self, settings):
        """ Restore the tool's state from the settings. """

        # This default implementation does nothing.
        pass

    def save_data(self):
        """ Save the tool's data and return True if there was no error. """

        # This default implementation does nothing.
        return True

    def save_state(self, settings):
        """ Save the tool's state in the settings. """

        # This default implementation does nothing.
        pass


class ShellTool(ActionTool):
    """ A base class for tools that appear in a shell. """

    @property
    @abstractmethod
    def location(self):
        """ Get the location of the tool in the shell. """

        ...

    @property
    @abstractmethod
    def title(self):
        """ Get the tool's title. """

        ...

    @property
    @abstractmethod
    def widget(self):
        """ Get the tool's widget. """

        ...



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QComboBox, QFormLayout

from .....models import VersionRange

from ....helpers import BaseDialog


class VersionsDialog(BaseDialog):
    """ This class implements the dialog for select the start and end versions.
    """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        project = self.shell.project

        form = QFormLayout()
        layout.addLayout(form)

        start_combo_box = QComboBox()
        end_combo_box = QComboBox()

        # The start list will have the list of versions with the first version
        # replaced by "First".  The end list will have the list of versions
        # excluding the first version and with "Latest" appended.
        for i, version in enumerate(project.versions):
            if i == 0:
                # The first version never actually appears itself.
                start_combo_box.addItem("First", '')
            else:
                start_combo_box.addItem(version, version)
                end_combo_box.addItem(version, version)

        end_combo_box.addItem("Latest", '')

        form.addRow("Starting version", start_combo_box)
        form.addRow("Ending version", end_combo_box)

        self._start_combo_box = start_combo_box
        self._end_combo_box = end_combo_box

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        # Note that we ignore (and eventually discard) anything other than the
        # first version range.

        if len(self.model.versions) == 0:
            api_start = ''
            api_end = ''
        else:
            api_start = self.model.versions[0].startversion
            api_end = self.model.versions[0].endversion

        if api_start == '':
            start_index = 0
        else:
            start_index = project.versions.index(api_start)

        if api_end == '':
            end_index = len(project.versions) - 1
        else:
            end_index = project.versions.index(api_end) - 1

        self._start_combo_box.setCurrentIndex(start_index)
        self._end_combo_box.setCurrentIndex(end_index)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        start_combo_box = self._start_combo_box
        end_combo_box = self._end_combo_box

        start_version = start_combo_box.itemData(start_combo_box.currentIndex())
        end_version = end_combo_box.itemData(end_combo_box.currentIndex())

        self.model.versions = [VersionRange(startversion=start_version, endversion=end_version)]

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QFormLayout, QLineEdit

from ...helpers import BaseDialog
from ...shell import EventType

from .helpers import validate_platform_name


class NewPlatformDialog(BaseDialog):
    """ This class implements the dialog for creating a new platform. """

    def populate(self, layout):
        """ Populate the dialog's layout. """

        form = QFormLayout()
        layout.addLayout(form)

        self._platform = QLineEdit()
        form.addRow("Platform name", self._platform)

    def get_fields(self):
        """ Update the project from the dialog's fields. """

        project = self.model

        platform = self._platform.text().strip()
        if not validate_platform_name(platform, project, self):
            return False

        project.platforms.append(platform)

        self.shell.notify(EventType.PLATFORM_ADD, platform)

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass


@dataclass
class HeaderFileVersion:
    """ This class implements a single version of a C/C++ .h file. """

    # The MD5 signature of the header file excluding any initial comments.
    md5: str = ''

    # This specifies if the header file needs parsing.
    parse: bool = False

    # The version of the header file.  This will be empty if the project does
    # not have any versions defined.
    version: str = ''



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..code_container import CodeContainer
from ..enum import Enum
from ..function import Function
from ..klass import Class
from ..manual_code import ManualCode
from ..namespace import Namespace
from ..opaque_class import OpaqueClass
from ..operator_function import OperatorFunction
from ..typedef import Typedef
from ..variable import Variable

from .adapt import adapt
from .base_adapter import AttributeType, BaseAdapter


class SipFileAdapter(BaseAdapter):
    """ This is the SipFile adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'exportedheadercode':   AttributeType.LITERAL,
        'exportedtypehintcode': AttributeType.LITERAL,
        'initcode':             AttributeType.LITERAL,
        'modulecode':           AttributeType.LITERAL,
        'moduleheadercode':     AttributeType.LITERAL,
        'name':                 AttributeType.STRING,
        'postinitcode':         AttributeType.LITERAL,
        'preinitcode':          AttributeType.LITERAL,
        'typehintcode':         AttributeType.LITERAL,
    }

    # The map of element tags and Code sub-class factories.
    _TAG_CODE_MAP = {
        'Class':            Class,
        'Enum':             Enum,
        'Function':         Function,
        'ManualCode':       ManualCode,
        'Namespace':        Namespace,
        'OpaqueClass':      OpaqueClass,
        'OperatorFunction': OperatorFunction,
        'Typedef':          Typedef,
        'Variable':         Variable,
    }

    def as_str(self):
        """ Return the standard string representation. """

        return self.model.name

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        super().load(element, project, ui)

        adapt(self.model, CodeContainer).load(self._TAG_CODE_MAP, element,
                project, ui)

        # Progress any UI for the load.
        if ui is not None:
            ui.load_step()

    def save(self, output):
        """ Save the model to an output file. """

        sip_file = self.model

        output.write(f'<SipFile name="{sip_file.name}"')
        adapt(sip_file, CodeContainer).save_attributes(output)
        output.write('>\n')

        output += 1
        adapt(sip_file, CodeContainer).save_subelements(output)
        self.save_literal('exportedheadercode', output)
        self.save_literal('moduleheadercode', output)
        self.save_literal('modulecode', output)
        self.save_literal('preinitcode', output)
        self.save_literal('initcode', output)
        self.save_literal('postinitcode', output)
        self.save_literal('exportedtypehintcode', output)
        self.save_literal('typehintcode', output)
        output -= 1

        output.write('</SipFile>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from ..access import Access
from ..annos import Annos
from ..code import Code

from .adapt import adapt
from .base_adapter import AttributeType, BaseApiAdapter


class DestructorAdapter(BaseApiAdapter):
    """ This is the Destructor adapter. """

    # The map of attribute names and types.
    ATTRIBUTE_TYPE_MAP = {
        'methcode': AttributeType.LITERAL,
        'name':     AttributeType.STRING,
        'virtcode': AttributeType.LITERAL,
        'virtual':  AttributeType.BOOL,
    }

    def __eq__(self, other):
        """ Compare for C/C++ equality. """

        dtor = self.model
        other_dtor = other.model

        if type(dtor) is not type(other_dtor):
            return False

        if dtor.access != other_dtor.access:
            return False

        if dtor.name != other_dtor.name:
            return False

        if dtor.virtual != other_dtor.virtual:
            return False

        return True

    def as_str(self):
        """ Return the standard string representation. """

        dtor = self.model

        s = '~' + dtor.name + '()' + adapt(dtor, Annos).as_str()

        if dtor.virtual:
            s = 'virtual ' + s

        return s

    def generate_sip(self, sip_file, output):
        """ Generate the .sip file content. """

        dtor = self.model

        nr_ends = self.version_start(output)

        output.write(self.as_str())
        output.write(';\n')

        output.write_code_directive('%MethodCode', dtor.methcode)
        output.write_code_directive('%VirtualCatcherCode', dtor.virtcode)

        self.version_end(nr_ends, output)

    def load(self, element, project, ui):
        """ Load the model from the XML element.  An optional user interface
        may be available to inform the user of progress.
        """

        super().load(element, project, ui)

        adapt(self.model, Code).load(element, project, ui)
        adapt(self.model, Access).load(element, project, ui)

    def save(self, output):
        """ Save the model to an output file. """

        dtor = self.model

        output.write('<Destructor')
        adapt(dtor, Code).save_attributes(output)
        adapt(dtor, Access).save_attributes(output)
        self.save_attribute('name', dtor.name, output)
        self.save_bool('virtual', output)
        output.write('>\n')

        output += 1
        self.save_literal('methcode', output)
        self.save_literal('virtcode', output)
        adapt(dtor, Code).save_subelements(output)
        adapt(dtor, Access).save_subelements(output)
        output -= 1

        output.write('</Destructor>\n')



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from PyQt6.QtWidgets import QComboBox, QGridLayout, QLabel

from ....helpers import BaseDialog


class FeaturesDialog(BaseDialog):
    """ This class implements the dialog for selecting a number of features.
    """

    FEATURE_VALUES = ("Unset", "Set", "Inverted")

    def populate(self, layout):
        """ Populate the dialog's layout. """

        project = self.shell.project

        grid = QGridLayout()
        layout.insertLayout(0, grid)

        all_features = project.features + project.externalfeatures
        self._features = []

        for row, feature in enumerate(all_features):
            grid.addWidget(QLabel(feature), row, 0)

            combo_box = QComboBox()
            combo_box.insertItems(0, self.FEATURE_VALUES)
            grid.addWidget(combo_box, row, 1)
            self._features.append((combo_box, feature))

    def set_fields(self):
        """ Set the dialog's fields from the API item. """

        for combo_box, feature in self._features:
            for f in self.model.features:
                if f.startswith('!') and feature == f[1:]:
                    index = 2
                elif feature == f:
                    index = 1
                else:
                    index = 0

                combo_box.setCurrentIndex(index)

    def get_fields(self):
        """ Update the API item from the dialog's fields. """

        features = []

        for combo_box, feature in self._features:
            index = combo_box.currentIndex()

            if index == 2:
                features.append('!' + feature)
            elif index == 1:
                features.append(feature)

        self.model.features = features

        return True



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from abc import ABC, abstractmethod


class AbstractProjectUi(ABC):
    """ This class encapsulates the UI-related methods supporting the loading
    of a project.
    """

    @abstractmethod
    def error_creating_file(self, title, text, detail):
        """ Called when there was an error when creating a file. """

        ...

    @abstractmethod
    def load_starting(self, project, nr_steps):
        """ Called to initialise the UI prior to loading the project that will
        take a specific number of steps.
        """

        ...

    @abstractmethod
    def load_step(self):
        """ Called to update the UI once the next step of loading the project
        has been completed.
        """

        ...

    @abstractmethod
    def update_project_format(self, root_element, from_version, to_version):
        """ Called to update the project from it's current major version before
        it is parsed.  Return True if the user didn't cancel.
        """

        ...

    @abstractmethod
    def warn_minor_version_update(self, from_version, to_version):
        """ Called to warn the user that the project will be updated to the
        current minor version if saved.
        """

        ...



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from .api_editor_tool import ApiEditorTool
from .features_tool import FeaturesTool
from .import_project_tool import ImportProjectTool
from .logger_tool import LoggerTool
from .modules_tool import ModulesTool
from .platforms_tool import PlatformsTool
from .scanner_tool import ScannerTool
from .versions_tool import VersionsTool



# SPDX-License-Identifier: BSD-2-Clause

# Copyright (c) 2024 Phil Thompson <phil@riverbankcomputing.com>


from dataclasses import dataclass

from .code_container import CodeContainer


@dataclass
class SipFile(CodeContainer):
    """ This class implements a .sip file. """

    # The optional %ExportedHeaderCode.
    exportedheadercode: str = ''

    # The optional %ExportedTypeHintCode.
    exportedtypehintcode: str = ''

    # The optional %InitialisationCode.
    initcode: str = ''

    # The optional %ModuleCode.
    modulecode: str = ''

    # The optional %ModuleHeaderCode.
    moduleheadercode: str = ''

    # The basename of the header file.  All header files that contribute to the
    # .sip file will have the same basename.  Note that there may not be any
    # record in the project of those contributing header files (or the
    # directories they were contained in).
    name: str = ''

    # The optional %PostInitialisationCode.
    postinitcode: str = ''

    # The optional %PreInitialisationCode.
    preinitcode: str = ''

    # The optional %TypeHintCode.
    typehintcode: str = ''
